#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#


class ErrorIndication(Exception):  # noqa: N818
    """SNMPv3 error-indication values."""

    def __init__(self, descr=None):
        """Create an error indication object."""
        self.__value = self.__descr = (
            self.__class__.__name__[0].lower() + self.__class__.__name__[1:]
        )
        if descr:
            self.__descr = descr

    def __eq__(self, other) -> bool:
        """
        Compare the instance's value with another value for equality.

        Args:
            other: The value to compare against the instance's value.

        Returns:
            bool: True if the instance's value is equal to the other value, False otherwise.
        """
        return self.__value == other

    def __ne__(self, other) -> bool:
        """
        Compare the instance's value with another value for inequality.

        Args:
            other: The value to compare against the instance's value.

        Returns:
            bool: True if the instance's value is not equal to the other value, False otherwise.
        """
        return self.__value != other

    def __lt__(self, other):
        """
        Compare the instance's value with another value for less than.

        Args:
            other: The value to compare against the instance's value.

        Returns:
            bool: True if the instance's value is less than the other value, False otherwise.
        """
        return self.__value < other

    def __le__(self, other):
        """
        Compare the instance's value with another value for less than or equal to.

        Args:
            other: The value to compare against the instance's value.

        Returns:
            bool: True if the instance's value is less than or equal to the other value, False otherwise.
        """
        return self.__value <= other

    def __gt__(self, other):
        """
        Compare the instance's value with another value for greater than.

        Args:
            other: The value to compare against the instance's value.

        Returns:
            bool: True if the instance's value is greater than the other value, False otherwise.
        """
        return self.__value > other

    def __ge__(self, other):
        """
        Compare the instance's value with another value for greater than or equal to.

        Args:
            other: The value to compare against the instance's value.

        Returns:
            bool: True if the instance's value is greater than or equal to the other value, False otherwise.
        """
        return self.__value >= other

    def __str__(self):
        """Return error indication as a string."""
        return self.__descr


# SNMP message processing errors


class SerializationError(ErrorIndication):
    """SNMP message serialization error."""

    pass


serializationError = SerializationError(  # noqa: N816
    "SNMP message serialization error"
)


class DeserializationError(ErrorIndication):
    """SNMP message deserialization error."""

    pass


deserializationError = DeserializationError(  # noqa: N816
    "SNMP message deserialization error"
)


class ParseError(DeserializationError):
    """SNMP message parsing error."""

    pass


parseError = ParseError("SNMP message deserialization error")  # noqa: N816


class UnsupportedMsgProcessingModel(ErrorIndication):  # noqa: N818
    """Unsupported SNMP message processing model."""

    pass


unsupportedMsgProcessingModel = UnsupportedMsgProcessingModel(  # noqa: N816
    "Unknown SNMP message processing model ID encountered"
)


class UnknownPDUHandler(ErrorIndication):  # noqa: N818
    """Unknown SNMP PDU handler."""

    pass


unknownPDUHandler = UnknownPDUHandler("Unhandled PDU type encountered")  # noqa: N816


class UnsupportedPDUtype(ErrorIndication):  # noqa: N818
    """Unsupported SNMP PDU type."""

    pass


unsupportedPDUtype = UnsupportedPDUtype(  # noqa: N816
    "Unsupported SNMP PDU type encountered"
)


class RequestTimedOut(ErrorIndication):  # noqa: N818
    """SNMP request timed out."""

    pass


requestTimedOut = RequestTimedOut(  # noqa: N816
    "No SNMP response received before timeout"
)


class EmptyResponse(ErrorIndication):  # noqa: N818
    """Empty SNMP response message."""

    pass


emptyResponse = EmptyResponse("Empty SNMP response message")  # noqa: N816


class NonReportable(ErrorIndication):  # noqa: N818
    """SNMP report PDU generation not attempted."""

    pass


nonReportable = NonReportable("Report PDU generation not attempted")  # noqa: N816


class DataMismatch(ErrorIndication):  # noqa: N818
    """SNMP request/response parameters mismatch."""

    pass


dataMismatch = DataMismatch("SNMP request/response parameters mismatched")  # noqa: N816


class EngineIDMismatch(ErrorIndication):  # noqa: N818
    """SNMP engine ID mismatch."""

    pass


engineIDMismatch = EngineIDMismatch("SNMP engine ID mismatch encountered")  # noqa: N816


class UnknownEngineID(ErrorIndication):  # noqa: N818
    """Unknown SNMP engine ID."""

    pass


unknownEngineID = UnknownEngineID("Unknown SNMP engine ID encountered")  # noqa: N816


class TooBig(ErrorIndication):  # noqa: N818
    """SNMP message too big."""

    pass


tooBig = TooBig("SNMP message will be too big")  # noqa: N816


class LoopTerminated(ErrorIndication):  # noqa: N818
    """SNMP entities talk terminated."""

    pass


loopTerminated = LoopTerminated("Infinite SNMP entities talk terminated")  # noqa: N816


class InvalidMsg(ErrorIndication):  # noqa: N818
    """Invalid SNMP message header parameters."""

    pass


invalidMsg = InvalidMsg(  # noqa: N816
    "Invalid SNMP message header parameters encountered"
)


# SNMP security modules errors


class UnknownCommunityName(ErrorIndication):  # noqa: N818
    """Unknown SNMP community name."""

    pass


unknownCommunityName = UnknownCommunityName(  # noqa: N816
    "Unknown SNMP community name encountered"
)


class NoEncryption(ErrorIndication):  # noqa: N818
    """No encryption services configured."""

    pass


noEncryption = NoEncryption("No encryption services configured")  # noqa: N816


class EncryptionError(ErrorIndication):
    """SNMP message encryption error."""

    pass


encryptionError = EncryptionError("Ciphering services not available")  # noqa: N816


class DecryptionError(ErrorIndication):
    """SNMP message decryption error."""

    pass


decryptionError = DecryptionError(  # noqa: N816
    "Ciphering services not available or ciphertext is broken"
)


class NoAuthentication(ErrorIndication):  # noqa: N818
    """No authentication services configured."""

    pass


noAuthentication = NoAuthentication(  # noqa: N816
    "No authentication services configured"
)


class AuthenticationError(ErrorIndication):
    """SNMP message authentication error."""

    pass


authenticationError = AuthenticationError(  # noqa: N816
    "Ciphering services not available or bad parameters"
)


class AuthenticationFailure(ErrorIndication):  # noqa: N818
    """SNMP message authentication failure."""

    pass


authenticationFailure = AuthenticationFailure("Authenticator mismatched")  # noqa: N816


class UnsupportedAuthProtocol(ErrorIndication):  # noqa: N818
    """Unsupported SNMP authentication protocol."""

    pass


unsupportedAuthProtocol = UnsupportedAuthProtocol(  # noqa: N816
    "Authentication protocol is not supported"
)


class UnsupportedPrivProtocol(ErrorIndication):  # noqa: N818
    """Unsupported SNMP privacy protocol."""

    pass


unsupportedPrivProtocol = UnsupportedPrivProtocol(  # noqa: N816
    "Privacy protocol is not supported"
)


class UnknownSecurityName(ErrorIndication):  # noqa: N818
    """Unknown SNMP security name."""

    pass


unknownSecurityName = UnknownSecurityName(  # noqa: N816
    "Unknown SNMP security name encountered"
)


class UnsupportedSecurityModel(ErrorIndication):  # noqa: N818
    """Unsupported SNMP security model."""

    pass


unsupportedSecurityModel = UnsupportedSecurityModel(  # noqa: N816
    "Unsupported SNMP security model"
)


class UnsupportedSecurityLevel(ErrorIndication):  # noqa: N818
    """Unsupported SNMP security level."""

    pass


unsupportedSecurityLevel = UnsupportedSecurityLevel(  # noqa: N816
    "Unsupported SNMP security level"
)


class NotInTimeWindow(ErrorIndication):  # noqa: N818
    """SNMP message timing parameters not in windows of trust."""

    pass


notInTimeWindow = NotInTimeWindow(  # noqa: N816
    "SNMP message timing parameters not in windows of trust"
)


class UnknownUserName(ErrorIndication):  # noqa: N818
    """Unknown SNMP user name."""

    pass


unknownUserName = UnknownUserName("Unknown USM user")  # noqa: N816


class WrongDigest(ErrorIndication):  # noqa: N818
    """Wrong SNMP PDU digest."""

    pass


wrongDigest = WrongDigest("Wrong SNMP PDU digest")  # noqa: N816


class ReportPduReceived(ErrorIndication):  # noqa: N818
    """Remote SNMP engine reported error."""

    pass


reportPduReceived = ReportPduReceived("Remote SNMP engine reported error")  # noqa: N816


# SNMP access-control errors


class NoSuchView(ErrorIndication):  # noqa: N818
    """No such MIB view currently exists."""

    pass


noSuchView = NoSuchView("No such MIB view currently exists")  # noqa: N816


class NoAccessEntry(ErrorIndication):  # noqa: N818
    """Access to MIB node denied."""

    pass


noAccessEntry = NoAccessEntry("Access to MIB node denied")  # noqa: N816


class NoGroupName(ErrorIndication):  # noqa: N818
    """No such VACM group configured."""

    pass


noGroupName = NoGroupName("No such VACM group configured")  # noqa: N816


class NoSuchContext(ErrorIndication):  # noqa: N818
    """No such SNMP context exists."""

    pass


noSuchContext = NoSuchContext("SNMP context now found")  # noqa: N816


class NotInView(ErrorIndication):  # noqa: N818
    """Requested OID is out of MIB view."""

    pass


notInView = NotInView("Requested OID is out of MIB view")  # noqa: N816


class AccessAllowed(ErrorIndication):  # noqa: N818
    """Access to MIB node allowed."""

    pass


accessAllowed = AccessAllowed()  # noqa: N816


class OtherError(ErrorIndication):
    """Unspecified SNMP engine error."""

    pass


otherError = OtherError("Unspecified SNMP engine error occurred")  # noqa: N816


# SNMP Apps errors


class OidNotIncreasing(ErrorIndication):  # noqa: N818
    """OID not increasing."""

    pass


oidNotIncreasing = OidNotIncreasing("OID not increasing")  # noqa: N816



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#

from typing import TYPE_CHECKING


from pysnmp.proto import error
from pysnmp.proto.errind import ErrorIndication
from pysnmp.proto.secmod import cache

if TYPE_CHECKING:
    from pysnmp.entity.engine import SnmpEngine


class AbstractSecurityModel:
    """Abstract security model class."""

    SECURITY_MODEL_ID = None
    _cache: cache.Cache

    def __init__(self):
        """Create a security model object."""
        self._cache = cache.Cache()

    def process_incoming_message(
        self,
        snmpEngine: "SnmpEngine",
        messageProcessingModel,
        maxMessageSize,
        securityParameters,
        securityModel,
        securityLevel,
        wholeMsg,
        msg,
    ):
        """Process an incoming message."""
        raise error.ProtocolError("Security model %s not implemented" % self)

    def generate_request_message(
        self,
        snmpEngine: "SnmpEngine",
        messageProcessingModel,
        globalData,
        maxMessageSize,
        securityModel,
        securityEngineID,
        securityName,
        securityLevel,
        scopedPDU,
    ):
        """Generate a request message."""
        raise error.ProtocolError("Security model %s not implemented" % self)

    def generate_response_message(
        self,
        snmpEngine: "SnmpEngine",
        messageProcessingModel,
        globalData,
        maxMessageSize,
        securityModel,
        securityEngineID,
        securityName,
        securityLevel,
        scopedPDU,
        securityStateReference,
        ctx: ErrorIndication,
    ):
        """Generate a response message."""
        raise error.ProtocolError("Security model %s not implemented" % self)

    def release_state_information(self, stateReference):
        """Release state information."""
        self._cache.pop(stateReference)

    def receive_timer_tick(self, snmpEngine: "SnmpEngine", timeNow):
        """Process a timer tick."""
        pass

    def _close(self):
        """
        Close the security model to test memory leak.

        This method is intended for unit testing purposes only.
        It closes the security model and checks if all associated resources are released.
        """
        raise error.ProtocolError("Security model %s not implemented" % self)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import warnings
from bisect import bisect


class OrderedDict(dict):
    """Ordered dictionary used for indices."""

    def __init__(self, *args, **kwargs):
        """Create an ordered dictionary."""
        self.__keys = []
        self.__dirty = True
        super().__init__()
        if args:
            self.update(*args)
        if kwargs:
            self.update(**kwargs)

    def __setitem__(self, key, value):
        """Set an item in the dictionary."""
        if key not in self:
            self.__keys.append(key)
            self.__dirty = True
        super().__setitem__(key, value)

    def __delitem__(self, key):
        """Delete an item from the dictionary."""
        if key in self:
            self.__keys.remove(key)
            self.__dirty = True
        super().__delitem__(key)

    def clear(self):
        """Clear the dictionary."""
        super().clear()
        self.__keys = []
        self.__dirty = True

    def keys(self):
        """Return the keys in the dictionary."""
        if self.__dirty:
            self.__order()
        return list(self.__keys)

    def values(self):
        """Return the values in the dictionary."""
        if self.__dirty:
            self.__order()
        return [self[k] for k in self.__keys]

    def items(self):
        """Return the items in the dictionary."""
        if self.__dirty:
            self.__order()
        return [(k, self[k]) for k in self.__keys]

    def update(self, *args, **kwargs):
        """Update the dictionary."""
        if args:
            iterable = args[0]
            if hasattr(iterable, "keys"):
                for k in iterable:
                    self[k] = iterable[k]
            else:
                for k, v in iterable:
                    self[k] = v

        if kwargs:
            for k in kwargs:
                self[k] = kwargs[k]

    def sorting_function(self, keys):
        """Sort the keys in the dictionary."""
        keys.sort()

    def __order(self):
        self.sorting_function(self.__keys)
        self.__keysLens = sorted({len(k) for k in self.__keys}, reverse=True)
        self.__dirty = False

    def next_key(self, key):
        """Return the next key in the dictionary."""
        if self.__dirty:
            self.__order()

        keys = self.__keys

        nextIdx = bisect(keys, key)

        if nextIdx < len(keys):
            return keys[nextIdx]

        else:
            raise KeyError(key)

    def get_keys_lengths(self):
        """Return the keys lengths in the dictionary."""
        if self.__dirty:
            self.__order()
        return self.__keysLens

    # Compatibility API
    # compatibility with legacy code
    # Old to new attribute mapping
    deprecated_attributes = {
        "nextKey": "next_key",
        "getKeysLens": "get_keys_lengths",
    }

    def __getattr__(self, attr: str):
        """Handle deprecated attributes."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )


class OidOrderedDict(OrderedDict):
    """OID-ordered dictionary used for indices."""

    def __init__(self, *args, **kwargs):
        """Create an OID-ordered dictionary."""
        self.__keysCache = {}
        OrderedDict.__init__(self, *args, **kwargs)

    def __setitem__(self, key, value):
        """Set an item in the dictionary."""
        OrderedDict.__setitem__(self, key, value)
        if key not in self.__keysCache:
            if isinstance(key, tuple):
                self.__keysCache[key] = key
            else:
                self.__keysCache[key] = [int(x) for x in key.split(".") if x]

    def __delitem__(self, key):
        """Delete an item from the dictionary."""
        OrderedDict.__delitem__(self, key)
        if key in self.__keysCache:
            del self.__keysCache[key]

    def sorting_function(self, keys):
        """Sort the keys in the dictionary."""
        keys.sort(key=lambda k, d=self.__keysCache: d[k])



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
"""
This module is part of the pysnmp library, providing high-level API (HLAPI) for SNMP operations using asyncio for asynchronous I/O.

The module imports various submodules from the `pysnmp.hlapi.v1arch.asyncio` package, which include:

- `auth`: Authentication mechanisms for SNMP.
- `cmdgen`: Command generator for SNMP operations.
- `dispatch`: Dispatcher for handling SNMP messages.
- `ntforg`: Notification originator for sending SNMP traps/informs.
- `slim`: Lightweight SNMP operations.
- `transport`: Transport layer for SNMP communication.

Additionally, it imports SNMP data types from `pysnmp.proto.rfc1902` and `pysnmp.smi.rfc1902`.
"""
from pysnmp.hlapi.v1arch.asyncio.auth import *
from pysnmp.hlapi.v1arch.asyncio.cmdgen import *
from pysnmp.hlapi.v1arch.asyncio.dispatch import *
from pysnmp.hlapi.v1arch.asyncio.ntforg import *
from pysnmp.hlapi.v1arch.asyncio.slim import *
from pysnmp.hlapi.v1arch.asyncio.transport import *
from pysnmp.proto.rfc1902 import *
from pysnmp.smi.rfc1902 import *



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import time

(MibScalarInstance,) = mibBuilder.import_symbols("SNMPv2-SMI", "MibScalarInstance")

(
    snmpEngineID,
    snmpEngineBoots,
    snmpEngineTime,
    snmpEngineMaxMessageSize,
) = mibBuilder.import_symbols(
    "SNMP-FRAMEWORK-MIB",
    "snmpEngineID",
    "snmpEngineBoots",
    "snmpEngineTime",
    "snmpEngineMaxMessageSize",
)

__snmpEngineID = MibScalarInstance(snmpEngineID.name, (0,), snmpEngineID.syntax)
__snmpEngineBoots = MibScalarInstance(
    snmpEngineBoots.name, (0,), snmpEngineBoots.syntax.clone(1)
)
__snmpEngineTime = MibScalarInstance(
    snmpEngineTime.name, (0,), snmpEngineTime.syntax.clone(int(time.time()))
)
__snmpEngineMaxMessageSize = MibScalarInstance(
    snmpEngineMaxMessageSize.name, (0,), snmpEngineMaxMessageSize.syntax.clone(4096)
)

mibBuilder.export_symbols(
    "__SNMP-FRAMEWORK-MIB",
    snmpEngineID=__snmpEngineID,
    snmpEngineBoots=__snmpEngineBoots,
    snmpEngineTime=__snmpEngineTime,
    snmpEngineMaxMessageSize=__snmpEngineMaxMessageSize,
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pyasn1.type import univ
from pysnmp import nextid
from pysnmp.proto import error, rfc1155, rfc1157

# Shortcuts to SNMP types
Integer = univ.Integer
OctetString = univ.OctetString
Null = univ.Null
null = Null("")
ObjectIdentifier = univ.ObjectIdentifier

IpAddress = rfc1155.IpAddress
NetworkAddress = rfc1155.NetworkAddress
Counter = rfc1155.Counter
Gauge = rfc1155.Gauge
TimeTicks = rfc1155.TimeTicks
Opaque = rfc1155.Opaque

VarBind = rfc1157.VarBind
VarBindList = rfc1157.VarBindList
GetRequestPDU = rfc1157.GetRequestPDU
GetNextRequestPDU = rfc1157.GetNextRequestPDU
GetResponsePDU = rfc1157.GetResponsePDU
SetRequestPDU = rfc1157.SetRequestPDU
TrapPDU = rfc1157.TrapPDU
Message = rfc1157.Message


class VarBindAPI:
    """Var-bind API."""

    @staticmethod
    def set_oid_value(varBind, oidVal):
        """Set OID and value components of var-bind."""
        oid, val = oidVal[0], oidVal[1]
        varBind.setComponentByPosition(0, oid)
        if val is None:
            val = null
        varBind.setComponentByPosition(1).getComponentByPosition(1).setComponentByType(
            val.getTagSet(),
            val,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
            innerFlag=True,
        )
        return varBind

    @staticmethod
    def get_oid_value(varBind):
        """Return OID and value components of var-bind."""
        return varBind[0], varBind[1].getComponent(1)


apiVarBind = VarBindAPI()  # noqa: N816

getNextRequestID = nextid.Integer(0xFFFFFF)  # noqa: N816


class PDUAPI:
    """SNMP PDU API."""

    _error_status = rfc1157.errorStatus.clone(0)
    _error_index = Integer(0)

    def set_defaults(self, pdu):
        """Set default values for SNMP PDU."""
        pdu.setComponentByPosition(
            0,
            getNextRequestID(),
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        pdu.setComponentByPosition(
            1,
            self._error_status,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        pdu.setComponentByPosition(
            2,
            self._error_index,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        varBindList = pdu.setComponentByPosition(3).getComponentByPosition(3)
        varBindList.clear()

    @staticmethod
    def get_request_id(pdu):
        """Return request ID component of SNMP PDU."""
        return pdu.getComponentByPosition(0)

    @staticmethod
    def set_request_id(pdu, value):
        """Set request ID component of SNMP PDU."""
        pdu.setComponentByPosition(0, value)

    @staticmethod
    def get_error_status(pdu):
        """Return error status component of SNMP PDU."""
        return pdu.getComponentByPosition(1)

    @staticmethod
    def set_error_status(pdu, value):
        """Set error status component of SNMP PDU."""
        pdu.setComponentByPosition(1, value)

    @staticmethod
    def get_error_index(pdu, muteErrors=False):
        """Return error index component of SNMP PDU."""
        errorIndex = pdu.getComponentByPosition(2)
        if errorIndex > len(pdu[3]):
            if muteErrors:
                return errorIndex.clone(len(pdu[3]))
            raise error.ProtocolError(
                f"Error index out of range: {errorIndex} > {len(pdu[3])}"
            )
        return errorIndex

    @staticmethod
    def set_error_index(pdu, value):
        """Set error index component of SNMP PDU."""
        pdu.setComponentByPosition(2, value)

    def set_end_of_mib_error(self, pdu, errorIndex):
        """Set end-of-MIB error status."""
        self.set_error_index(pdu, errorIndex)
        self.set_error_status(pdu, 2)

    def set_no_such_instance_error(self, pdu, errorIndex):
        """Set no-such-instance error status."""
        self.set_end_of_mib_error(pdu, errorIndex)

    @staticmethod
    def get_varbind_list(pdu):
        """Return var-bind list component of SNMP PDU."""
        return pdu.getComponentByPosition(3)

    @staticmethod
    def set_varbind_list(pdu, varBindList):
        """Set var-bind list component of SNMP PDU."""
        pdu.setComponentByPosition(3, varBindList)

    @staticmethod
    def get_varbinds(pdu):
        """Return var-binds component of SNMP PDU."""
        return [
            apiVarBind.get_oid_value(varBind)
            for varBind in pdu.getComponentByPosition(3)
        ]

    @staticmethod
    def set_varbinds(pdu, varBinds):
        """Set var-binds component of SNMP PDU."""
        varBindList = pdu.setComponentByPosition(3).getComponentByPosition(3)
        varBindList.clear()
        for idx, varBind in enumerate(varBinds):
            if isinstance(varBind, VarBind):
                varBindList.setComponentByPosition(idx, varBind)
            else:
                varBindList.setComponentByPosition(idx)
                apiVarBind.set_oid_value(
                    varBindList.getComponentByPosition(idx), varBind
                )

    def get_response(self, reqPDU):
        """Build response PDU."""
        rspPDU = GetResponsePDU()
        self.set_defaults(rspPDU)
        self.set_request_id(rspPDU, self.get_request_id(reqPDU))
        return rspPDU

    def get_varbind_table(self, reqPDU, rspPDU):
        """Return var-bind table."""
        if apiPDU.get_error_status(rspPDU) == 2:
            varBindRow = [(vb[0], null) for vb in apiPDU.get_varbinds(reqPDU)]
        else:
            varBindRow = apiPDU.get_varbinds(rspPDU)
        return [varBindRow]

    def get_next_varbinds(self, varBinds, errorIndex=None):
        """Return next var-binds."""
        errorIndication = None

        if errorIndex:
            return errorIndication, []

        rspVarBinds = [(vb[0], null) for vb in varBinds]

        return errorIndication, rspVarBinds


apiPDU = PDUAPI()  # noqa: N816


class TrapPDUAPI:
    """SNMP trap PDU API."""

    _networkAddress = None  # noqa: N815
    _entOid = ObjectIdentifier((1, 3, 6, 1, 4, 1, 20408))  # noqa: N815
    _genericTrap = rfc1157.genericTrap.clone("coldStart")  # noqa: N815
    _zeroInt = univ.Integer(0)  # noqa: N815
    _zeroTime = TimeTicks(0)  # noqa: N815

    def set_defaults(self, pdu):
        """Set default values for SNMP trap PDU."""
        if self._networkAddress is None:
            try:
                import socket

                agentAddress = IpAddress(socket.gethostbyname(socket.gethostname()))
            except Exception:
                agentAddress = IpAddress("0.0.0.0")
            self._networkAddress = NetworkAddress().setComponentByPosition(
                0, agentAddress
            )
        pdu.setComponentByPosition(
            0,
            self._entOid,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        pdu.setComponentByPosition(
            1,
            self._networkAddress,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        pdu.setComponentByPosition(
            2,
            self._genericTrap,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        pdu.setComponentByPosition(
            3,
            self._zeroInt,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        pdu.setComponentByPosition(
            4,
            self._zeroTime,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        varBindList = pdu.setComponentByPosition(5).getComponentByPosition(5)
        varBindList.clear()

    @staticmethod
    def get_enterprise(pdu):
        """Return enterprise component of SNMP trap PDU."""
        return pdu.getComponentByPosition(0)

    @staticmethod
    def set_enterprise(pdu, value):
        """Set enterprise component of SNMP trap PDU."""
        pdu.setComponentByPosition(0, value)

    @staticmethod
    def get_agent_address(pdu):
        """Return agent address component of SNMP trap PDU."""
        return pdu.getComponentByPosition(1).getComponentByPosition(0)

    @staticmethod
    def set_agent_address(pdu, value):
        """Set agent address component of SNMP trap PDU."""
        pdu.setComponentByPosition(1).getComponentByPosition(1).setComponentByPosition(
            0, value
        )

    @staticmethod
    def get_generic_trap(pdu):
        """Return generic trap component of SNMP trap PDU."""
        return pdu.getComponentByPosition(2)

    @staticmethod
    def set_generic_trap(pdu, value):
        """Set generic trap component of SNMP trap PDU."""
        pdu.setComponentByPosition(2, value)

    @staticmethod
    def get_specific_trap(pdu):
        """Return specific trap component of SNMP trap PDU."""
        return pdu.getComponentByPosition(3)

    @staticmethod
    def set_specific_trap(pdu, value):
        """Set specific trap component of SNMP trap PDU."""
        pdu.setComponentByPosition(3, value)

    @staticmethod
    def get_timestamp(pdu):
        """Return time stamp component of SNMP trap PDU."""
        return pdu.getComponentByPosition(4)

    @staticmethod
    def set_timestamp(pdu, value):
        """Set time stamp component of SNMP trap PDU."""
        pdu.setComponentByPosition(4, value)

    @staticmethod
    def get_varbind_list(pdu):
        """Return var-bind list component of SNMP trap PDU."""
        return pdu.getComponentByPosition(5)

    @staticmethod
    def set_varbind_list(pdu, varBindList):
        """Set var-bind list component of SNMP trap PDU."""
        pdu.setComponentByPosition(5, varBindList)

    @staticmethod
    def get_varbinds(pdu):
        """Return var-binds component of SNMP trap PDU."""
        varBinds = []
        for varBind in pdu.getComponentByPosition(5):
            varBinds.append(apiVarBind.get_oid_value(varBind))
        return varBinds

    @staticmethod
    def set_varbinds(pdu, varBinds):
        """Set var-binds component of SNMP trap PDU."""
        varBindList = pdu.setComponentByPosition(5).getComponentByPosition(5)
        varBindList.clear()
        idx = 0
        for varBind in varBinds:
            if isinstance(varBind, VarBind):
                varBindList.setComponentByPosition(idx, varBind)
            else:
                varBindList.setComponentByPosition(idx)
                apiVarBind.set_oid_value(
                    varBindList.getComponentByPosition(idx), varBind
                )
            idx += 1


apiTrapPDU = TrapPDUAPI()  # noqa: N816


class MessageAPI:
    """SNMP message API."""

    _version = rfc1157.version.clone(0)
    _community = univ.OctetString("public")

    def set_defaults(self, msg):
        """Set default values for SNMP message."""
        msg.setComponentByPosition(
            0,
            self._version,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        msg.setComponentByPosition(
            1,
            self._community,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        return msg

    @staticmethod
    def get_version(msg):
        """Return version component of SNMP message."""
        return msg.getComponentByPosition(0)

    @staticmethod
    def set_version(msg, value):
        """Set version component of SNMP message."""
        msg.setComponentByPosition(0, value)

    @staticmethod
    def get_community(msg):
        """Return community component of SNMP message."""
        return msg.getComponentByPosition(1)

    @staticmethod
    def set_community(msg, value):
        """Set community component of SNMP message."""
        msg.setComponentByPosition(1, value)

    @staticmethod
    def get_pdu(
        msg,
    ) -> "TrapPDU | GetRequestPDU | GetNextRequestPDU | GetResponsePDU | SetRequestPDU":
        """Return PDU component of SNMP message."""
        return msg.getComponentByPosition(2).getComponent()

    @staticmethod
    def set_pdu(msg, value):
        """Set PDU component of SNMP message."""
        msg.setComponentByPosition(2).getComponentByPosition(2).setComponentByType(
            value.getTagSet(),
            value,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
            innerFlag=True,
        )

    def get_response(self, reqMsg):
        """Return a response message to a request message."""
        rspMsg = Message()
        self.set_defaults(rspMsg)
        self.set_version(rspMsg, self.get_version(reqMsg))
        self.set_community(rspMsg, self.get_community(reqMsg))
        self.set_pdu(rspMsg, apiPDU.get_response(self.get_pdu(reqMsg)))
        return rspMsg


apiMessage = MessageAPI()  # noqa: N816



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from hashlib import md5

from pyasn1.type import univ
from pysnmp.proto import errind, error
from pysnmp.proto.secmod.rfc3414 import localkey
from pysnmp.proto.secmod.rfc3414.auth import base

TWELVE_ZEROS = univ.OctetString((0,) * 12).asOctets()
FORTY_EIGHT_ZEROS = (0,) * 48


# rfc3414: 6.2.4


class HmacMd5(base.AbstractAuthenticationService):
    """The HMAC-MD5-96 authentication protocol.

    https://tools.ietf.org/html/rfc3414#section-6.2.4
    """

    SERVICE_ID = (1, 3, 6, 1, 6, 3, 10, 1, 1, 2)  # usmHMACMD5AuthProtocol
    IPAD = [0x36] * 64
    OPAD = [0x5C] * 64

    def hash_passphrase(self, authKey):
        """Hash a passphrase."""
        return localkey.hash_passphrase_md5(authKey)

    def localize_key(self, authKey, snmpEngineID):
        """Localize a key."""
        return localkey.localize_key_md5(authKey, snmpEngineID)

    @property
    def digest_length(self):
        """Return digest length."""
        return 12

    # 6.3.1
    def authenticate_outgoing_message(self, authKey, wholeMsg):
        """Authenticate outgoing message."""
        # Here we expect calling secmod to indicate where the digest
        # should be in the substrate. Also, it pre-sets digest placeholder
        # so we hash wholeMsg out of the box.
        # Yes, that's ugly but that's rfc...
        value = wholeMsg.find(TWELVE_ZEROS)
        if value == -1:
            raise error.ProtocolError("Cant locate digest placeholder")
        wholeHead = wholeMsg[:value]
        wholeTail = wholeMsg[value + 12 :]

        # 6.3.1.1

        # 6.3.1.2a
        extendedAuthKey = authKey.asNumbers() + FORTY_EIGHT_ZEROS

        # 6.3.1.2b --> no-op

        # 6.3.1.2c
        k1 = univ.OctetString(map(lambda x, y: x ^ y, extendedAuthKey, self.IPAD))

        # 6.3.1.2d --> no-op

        # 6.3.1.2e
        k2 = univ.OctetString(map(lambda x, y: x ^ y, extendedAuthKey, self.OPAD))

        # 6.3.1.3
        # noinspection PyDeprecation,PyCallingNonCallable
        d1 = md5(k1.asOctets() + wholeMsg).digest()

        # 6.3.1.4
        # noinspection PyDeprecation,PyCallingNonCallable
        d2 = md5(k2.asOctets() + d1).digest()
        mac = d2[:12]

        # 6.3.1.5 & 6
        return wholeHead + mac + wholeTail

    # 6.3.2
    def authenticate_incoming_message(self, authKey, authParameters, wholeMsg):
        """Authenticate incoming message."""
        # 6.3.2.1 & 2
        if len(authParameters) != 12:
            raise error.StatusInformation(errorIndication=errind.authenticationError)

        # 6.3.2.3
        value = wholeMsg.find(authParameters.asOctets())
        if value == -1:
            raise error.ProtocolError("Cant locate digest in wholeMsg")
        wholeHead = wholeMsg[:value]
        wholeTail = wholeMsg[value + 12 :]
        authenticatedWholeMsg = wholeHead + TWELVE_ZEROS + wholeTail

        # 6.3.2.4a
        extendedAuthKey = authKey.asNumbers() + FORTY_EIGHT_ZEROS

        # 6.3.2.4b --> no-op

        # 6.3.2.4c
        k1 = univ.OctetString(map(lambda x, y: x ^ y, extendedAuthKey, self.IPAD))

        # 6.3.2.4d --> no-op

        # 6.3.2.4e
        k2 = univ.OctetString(map(lambda x, y: x ^ y, extendedAuthKey, self.OPAD))

        # 6.3.2.5a
        # noinspection PyDeprecation,PyCallingNonCallable
        d1 = md5(k1.asOctets() + authenticatedWholeMsg).digest()

        # 6.3.2.5b
        # noinspection PyDeprecation,PyCallingNonCallable
        d2 = md5(k2.asOctets() + d1).digest()

        # 6.3.2.5c
        mac = d2[:12]

        # 6.3.2.6
        if mac != authParameters:
            raise error.StatusInformation(errorIndication=errind.authenticationFailure)

        return authenticatedWholeMsg



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import warnings

from pysnmp import debug
from pysnmp.smi import error
from pysnmp.smi.builder import MibBuilder
from pysnmp.smi.indices import OidOrderedDict, OrderedDict

__all__ = ["MibViewController"]

classTypes = (type,)  # noqa: N816
instanceTypes = (object,)  # noqa: N816


class MibViewController:
    """Create a MIB view controller."""

    def __init__(self, mibBuilder: MibBuilder):
        """Create a MIB view controller."""
        self.mibBuilder = mibBuilder
        self.lastBuildId = -1
        self.__mibSymbolsIdx = OrderedDict()

    # Indexing part

    def index_mib(self):
        """Re-index MIB view."""
        if self.lastBuildId == self.mibBuilder.lastBuildId:
            return

        debug.logger & debug.FLAG_MIB and debug.logger("indexMib: re-indexing MIB view")

        (MibScalarInstance,) = self.mibBuilder.import_symbols(  # type: ignore
            "SNMPv2-SMI", "MibScalarInstance"
        )

        #
        # Create indices
        #

        # Module name -> module-scope indices
        self.__mibSymbolsIdx.clear()

        # Oid <-> label indices

        # This is potentially ambiguous mapping. Sort modules in
        # ascending age for resolution
        def __sorting_function(x, b: MibBuilder = self.mibBuilder):
            if b.module_id in b.mibSymbols[x]:
                m = b.mibSymbols[x][b.module_id]
                r = m.getRevisions()
                if r:
                    return r[0][0]

            return "1970-01-01 00:00"

        modNames = list(self.mibBuilder.mibSymbols.keys())
        modNames.sort(key=__sorting_function)

        # Index modules names
        for modName in [""] + modNames:
            # Modules index
            self.__mibSymbolsIdx[modName] = mibMod = {
                "oidToLabelIdx": OidOrderedDict(),
                "labelToOidIdx": {},
                "varToNameIdx": {},
                "typeToModIdx": OrderedDict(),
                "oidToModIdx": {},
            }

            if not modName:
                globMibMod = mibMod
                continue

            # Types & MIB vars indices
            for n, v in self.mibBuilder.mibSymbols[modName].items():
                if n == self.mibBuilder.module_id:  # do not index this
                    continue  # special symbol
                if isinstance(v, classTypes):
                    if n in mibMod["typeToModIdx"]:
                        raise error.SmiError(
                            "Duplicate SMI type {}::{}, has {}".format(
                                modName, n, mibMod["typeToModIdx"][n]
                            )
                        )
                    globMibMod["typeToModIdx"][n] = modName
                    mibMod["typeToModIdx"][n] = modName
                elif isinstance(v, instanceTypes):
                    if isinstance(v, MibScalarInstance):
                        continue
                    if n in mibMod["varToNameIdx"]:
                        raise error.SmiError(
                            "Duplicate MIB variable {}::{} has {}".format(
                                modName, n, mibMod["varToNameIdx"][n]
                            )
                        )
                    globMibMod["varToNameIdx"][n] = v.name
                    mibMod["varToNameIdx"][n] = v.name
                    # Potentionally ambiguous mapping ahead
                    globMibMod["oidToModIdx"][v.name] = modName
                    mibMod["oidToModIdx"][v.name] = modName
                    globMibMod["oidToLabelIdx"][v.name] = (n,)
                    mibMod["oidToLabelIdx"][v.name] = (n,)
                else:
                    raise error.SmiError(f"Unexpected object {modName}::{n}")

        # Build oid->long-label index
        oidToLabelIdx = self.__mibSymbolsIdx[""]["oidToLabelIdx"]
        labelToOidIdx = self.__mibSymbolsIdx[""]["labelToOidIdx"]
        prevOid = ()
        baseLabel = ()
        for key in oidToLabelIdx.keys():
            keydiff = len(key) - len(prevOid)
            if keydiff > 0:
                if prevOid:
                    if keydiff == 1:
                        baseLabel = oidToLabelIdx[prevOid]
                    else:
                        baseLabel += key[-keydiff:-1]
                else:
                    baseLabel = ()
            elif keydiff < 0:
                baseLabel = ()
                keyLen = len(key)
                i = keyLen - 1
                while i:
                    k = key[:i]
                    if k in oidToLabelIdx:
                        baseLabel = oidToLabelIdx[k]
                        if i != keyLen - 1:
                            baseLabel += key[i:-1]
                        break
                    i -= 1
            # Build oid->long-label index
            oidToLabelIdx[key] = baseLabel + oidToLabelIdx[key]
            # Build label->oid index
            labelToOidIdx[oidToLabelIdx[key]] = key
            prevOid = key

        # Build module-scope oid->long-label index
        for mibMod in self.__mibSymbolsIdx.values():
            for oid in mibMod["oidToLabelIdx"].keys():
                mibMod["oidToLabelIdx"][oid] = oidToLabelIdx[oid]
                mibMod["labelToOidIdx"][oidToLabelIdx[oid]] = oid

        self.lastBuildId = self.mibBuilder.lastBuildId

    # Module management

    def get_ordered_module_name(self, index):
        """Return module name by index."""
        self.index_mib()
        modNames = self.__mibSymbolsIdx.keys()
        if modNames:
            return modNames[index]
        raise error.SmiError("No modules loaded at %s" % self)

    def get_first_module_name(self):
        """Return first module name."""
        return self.get_ordered_module_name(0)

    def get_last_module_name(self):
        """Return last module name."""
        return self.get_ordered_module_name(-1)

    def get_next_module_name(self, modName):
        """Return next module name."""
        self.index_mib()
        try:
            return self.__mibSymbolsIdx.next_key(modName)
        except KeyError:
            raise error.SmiError(f"No module next to {modName} at {self}")

    # MIB tree node management

    def __get_oid_label(self, nodeName, oidToLabelIdx, labelToOidIdx):
        """getOidLabel(nodeName) -> (oid, label, suffix)."""
        if not nodeName:
            return nodeName, nodeName, ()
        if nodeName in labelToOidIdx:
            return labelToOidIdx[nodeName], nodeName, ()
        if nodeName in oidToLabelIdx:
            return nodeName, oidToLabelIdx[nodeName], ()
        if len(nodeName) < 2:
            return nodeName, nodeName, ()
        oid, label, suffix = self.__get_oid_label(
            nodeName[:-1], oidToLabelIdx, labelToOidIdx
        )
        suffix = suffix + nodeName[-1:]
        resLabel = label + tuple(str(x) for x in suffix)
        if resLabel in labelToOidIdx:
            return labelToOidIdx[resLabel], resLabel, ()
        resOid = oid + suffix
        if resOid in oidToLabelIdx:
            return resOid, oidToLabelIdx[resOid], ()
        return oid, label, suffix

    def get_node_name_by_oid(self, nodeName, modName=""):
        """Return node name by OID."""
        self.index_mib()
        if modName in self.__mibSymbolsIdx:
            mibMod = self.__mibSymbolsIdx[modName]
        else:
            raise error.SmiError(f"No module {modName} at {self}")
        oid, label, suffix = self.__get_oid_label(
            nodeName, mibMod["oidToLabelIdx"], mibMod["labelToOidIdx"]
        )
        if oid == label:
            raise error.NoSuchObjectError(
                str=f"Can't resolve node name {modName}::{nodeName} at {self}"
            )
        debug.logger & debug.FLAG_MIB and debug.logger(
            f"getNodeNameByOid: resolved {modName}:{nodeName} -> {label}.{suffix}"
        )
        return oid, label, suffix

    def get_node_name_by_desc(self, nodeName, modName=""):
        """Return node name by MIB symbol."""
        self.index_mib()
        if modName in self.__mibSymbolsIdx:
            mibMod = self.__mibSymbolsIdx[modName]
        else:
            raise error.SmiError(f"No module {modName} at {self}")
        if nodeName in mibMod["varToNameIdx"]:
            oid = mibMod["varToNameIdx"][nodeName]
        else:
            raise error.NoSuchObjectError(
                str=f"No such symbol {modName}::{nodeName} at {self}"
            )
        debug.logger & debug.FLAG_MIB and debug.logger(
            f"getNodeNameByDesc: resolved {modName}:{nodeName} -> {oid}"
        )
        return self.get_node_name_by_oid(oid, modName)

    def get_node_name(self, nodeName, modName=""):
        """Return node name by OID or MIB symbol."""
        # nodeName may be either an absolute OID/label or a
        # ( MIB-symbol, su, ff, ix)
        try:
            # First try nodeName as an OID/label
            return self.get_node_name_by_oid(nodeName, modName)
        except error.NoSuchObjectError:
            # ...on failure, try as MIB symbol
            oid, label, suffix = self.get_node_name_by_desc(nodeName[0], modName)
            # ...with trailing suffix
            return self.get_node_name_by_oid(oid + suffix + nodeName[1:], modName)

    def get_ordered_node_name(self, index, modName=""):
        """Return node name by index."""
        self.index_mib()
        if modName in self.__mibSymbolsIdx:
            mibMod = self.__mibSymbolsIdx[modName]
        else:
            raise error.SmiError(f"No module {modName} at {self}")
        if not mibMod["oidToLabelIdx"]:
            raise error.NoSuchObjectError(
                str=f"No variables at MIB module {modName} at {self}"
            )
        try:
            oid, label = mibMod["oidToLabelIdx"].items()[index]
        except KeyError:
            raise error.NoSuchObjectError(
                str=f"No symbol at position {index} in MIB module {modName} at {self}"
            )
        return oid, label, ()

    def get_first_node_name(self, modName=""):
        """Return first node name."""
        return self.get_ordered_node_name(0, modName)

    def get_last_node_name(self, modName=""):
        """Return last node name."""
        return self.get_ordered_node_name(-1, modName)

    def get_next_node_name(self, nodeName, modName=""):
        """Return next node name."""
        oid, label, suffix = self.get_node_name(nodeName, modName)
        try:
            return self.get_node_name(
                self.__mibSymbolsIdx[modName]["oidToLabelIdx"].next_key(oid) + suffix,
                modName,
            )
        except KeyError:
            raise error.NoSuchObjectError(
                str=f"No name next to {modName}::{nodeName} at {self}"
            )

    def get_parent_node_name(self, nodeName, modName=""):
        """Return parent node name."""
        oid, label, suffix = self.get_node_name(nodeName, modName)
        if len(oid) < 2:
            raise error.NoSuchObjectError(
                str=f"No parent name for {modName}::{nodeName} at {self}"
            )
        return oid[:-1], label[:-1], oid[-1:] + suffix

    def get_node_location(self, nodeName, modName=""):
        """Return module name, node name, and suffix."""
        oid, label, suffix = self.get_node_name(nodeName, modName)
        return self.__mibSymbolsIdx[""]["oidToModIdx"][oid], label[-1], suffix

    # MIB type management

    def get_type_name(self, typeName, modName=""):
        """Return type name by MIB symbol."""
        self.index_mib()
        if modName in self.__mibSymbolsIdx:
            mibMod = self.__mibSymbolsIdx[modName]
        else:
            raise error.SmiError(f"No module {modName} at {self}")
        if typeName in mibMod["typeToModIdx"]:
            m = mibMod["typeToModIdx"][typeName]
        else:
            raise error.NoSuchObjectError(
                str=f"No such type {modName}::{typeName} at {self}"
            )
        return m, typeName

    def get_ordered_type_name(self, index, modName=""):
        """Return type name by index."""
        self.index_mib()
        if modName in self.__mibSymbolsIdx:
            mibMod = self.__mibSymbolsIdx[modName]
        else:
            raise error.SmiError(f"No module {modName} at {self}")
        if not mibMod["typeToModIdx"]:
            raise error.NoSuchObjectError(
                str=f"No types at MIB module {modName} at {self}"
            )
        t = mibMod["typeToModIdx"].keys()[index]
        return mibMod["typeToModIdx"][t], t

    def get_first_type_name(self, modName=""):
        """Get first type name."""
        return self.get_ordered_type_name(0, modName)

    def get_last_type_name(self, modName=""):
        """Get last type name."""
        return self.get_ordered_type_name(-1, modName)

    def get_next_type(self, typeName, modName=""):
        """Get next type name."""
        m, t = self.get_type_name(typeName, modName)
        try:
            return self.__mibSymbolsIdx[m]["typeToModIdx"].next_key(t)
        except KeyError:
            raise error.NoSuchObjectError(
                str=f"No type next to {modName}::{typeName} at {self}"
            )

    # Compatibility API
    deprecated_attributes = {
        "getOrderedModuleName": "get_ordered_module_name",
        "getFirstModuleName": "get_first_module_name",
        "getLastModuleName": "get_last_module_name",
        "getNextModuleName": "get_next_module_name",
        "getNodeNameByOid": "get_node_name_by_oid",
        "getNodeNameByDesc": "get_node_name_by_desc",
        "getNodeName": "get_node_name",
        "getOrderedNodeName": "get_ordered_node_name",
        "getFirstNodeName": "get_first_node_name",
        "getLastNodeName": "get_last_node_name",
        "getNextNodeName": "get_next_node_name",
        "getParentNodeName": "get_parent_node_name",
        "getNodeLocation": "get_node_location",
        "getTypeName": "get_type_name",
        "getOrderedTypeName": "get_ordered_type_name",
        "getFirstTypeName": "get_first_type_name",
        "getLastTypeName": "get_last_type_name",
        "getNextType": "get_next_type",
    }

    def __getattr__(self, attr):
        """Redirect some attrs access to the OID object to behave alike."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pyasn1.error import PyAsn1Error
from pyasn1.type import constraint, namedtype, tag, univ
from pysnmp.proto import error
from pysnmp.smi.error import SmiError

__all__ = [
    "Opaque",
    "NetworkAddress",
    "ObjectName",
    "TimeTicks",
    "Counter",
    "Gauge",
    "IpAddress",
]


class IpAddress(univ.OctetString):
    """IP address."""

    tagSet = univ.OctetString.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x00)
    )
    subtypeSpec = univ.OctetString.subtypeSpec + constraint.ValueSizeConstraint(
        4, 4
    )  # noqa: N815

    def prettyIn(self, value):  # noqa: N802
        """Return a pretty-formatted IP address."""
        if isinstance(value, str) and len(value) != 4:
            try:
                value = [int(x) for x in value.split(".")]
            except:
                raise error.ProtocolError("Bad IP address syntax %s" % value)
        if len(value) != 4:
            raise error.ProtocolError("Bad IP address syntax")
        return univ.OctetString.prettyIn(self, value)

    def prettyOut(self, value):  # noqa: N802
        """Return a pretty-printed IP address."""
        if value:
            return ".".join(["%d" % x for x in self.__class__(value).asNumbers()])
        else:
            return ""


class Counter(univ.Integer):
    """Counter."""

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x01)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class NetworkAddress(univ.Choice):
    """Network address."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("internet", IpAddress())
    )

    def clone(self, value=univ.noValue, **kwargs):
        """Clone this instance.

        If *value* is specified, use its tag as the component type selector,
        and itself as the component value.

        :param value: (Optional) the component value.
        :type value: :py:obj:`pyasn1.type.base.Asn1ItemBase`
        :return: the cloned instance.
        :rtype: :py:obj:`pysnmp.proto.rfc1155.NetworkAddress`
        :raise: :py:obj:`pysnmp.smi.error.SmiError`:
            if the type of *value* is not allowed for this Choice instance.
        """
        cloned = univ.Choice.clone(self, **kwargs)
        if value is not univ.noValue:
            if isinstance(value, NetworkAddress):
                value = value.getComponent()
            elif not isinstance(value, IpAddress):
                # IpAddress is the only supported type, perhaps forever because
                # this is SNMPv1.
                value = IpAddress(value)
            try:
                tagSet = value.tagSet
            except AttributeError:
                raise PyAsn1Error(f"component value {value!r} has no tag set")
            cloned.setComponentByType(tagSet, value)
        return cloned

    # RFC 1212, section 4.1.6:
    #
    #    "(5)  NetworkAddress-valued: `n+1' sub-identifiers, where `n'
    #          depends on the kind of address being encoded (the first
    #          sub-identifier indicates the kind of address, value 1
    #          indicates an IpAddress);"

    def clone_from_name(self, value, impliedFlag, parentRow, parentIndices):
        """Clone this instance from a tuple of sub-identifiers."""
        kind = value[0]
        clone = self.clone()
        if kind == 1:
            clone["internet"] = tuple(value[1:5])
            return clone, value[5:]
        else:
            raise SmiError(f"unknown NetworkAddress type {kind!r}")

    def clone_as_name(self, impliedFlag, parentRow, parentIndices):
        """Return a tuple of sub-identifiers representing this instance.

        :param bool impliedFlag: (Unused) whether the value is implied.
        :param parentRow: (Unused) the parent row.
        :param parentIndices: (Unused) the parent indices.
        :return: the sub-identifiers.
        :rtype: :py:obj:`tuple`
        :raise: :py:obj:`pysnmp.smi.error.SmiError`:
            if the type of the component value is not allowed for this Choice instance.
        """
        kind = self.getName()
        component = self.getComponent()
        if kind == "internet":
            return (1,) + tuple(component.asNumbers())
        else:
            raise SmiError(f"unknown NetworkAddress type {kind!r}")


class Gauge(univ.Integer):
    """Gauge."""

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x02)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class TimeTicks(univ.Integer):
    """TimeTicks."""

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x03)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class Opaque(univ.OctetString):
    """Opaque."""

    tagSet = univ.OctetString.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x04)
    )


class ObjectName(univ.ObjectIdentifier):
    """Object name."""

    pass


class TypeCoercionHackMixIn:  # XXX keep this old-style class till pyasn1 types becomes new-style
    # Reduce ASN1 type check to simple tag check as SMIv2 objects may
    # not be constraints-compatible with those used in SNMP PDU.
    def _verify_component(self, idx, value, **kwargs):
        componentType = self._componentType  # noqa: N806
        if componentType:
            if idx >= len(componentType):
                raise PyAsn1Error("Component type error out of range")
            t = componentType[idx].getType()
            if not t.getTagSet().isSuperTagSetOf(value.getTagSet()):
                raise PyAsn1Error(f"Component type error {t!r} vs {value!r}")


class SimpleSyntax(TypeCoercionHackMixIn, univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("number", univ.Integer()),
        namedtype.NamedType("string", univ.OctetString()),
        namedtype.NamedType("object", univ.ObjectIdentifier()),
        namedtype.NamedType("empty", univ.Null()),
    )


class ApplicationSyntax(TypeCoercionHackMixIn, univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("address", NetworkAddress()),
        namedtype.NamedType("counter", Counter()),
        namedtype.NamedType("gauge", Gauge()),
        namedtype.NamedType("ticks", TimeTicks()),
        namedtype.NamedType("arbitrary", Opaque()),
    )


class ObjectSyntax(univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("simple", SimpleSyntax()),
        namedtype.NamedType("application-wide", ApplicationSyntax()),
    )



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
"""
This module initializes the pysnmp high-level API (HLAPI) for SNMPv1 architecture.

It imports various components from the asyncio-based API, including authentication,
dispatch, and protocol modules. Additionally, it imports specific SNMP protocol
objects from RFC1902 and RFC1905.

Modules and objects imported:
- `pysnmp.hlapi.v1arch.asyncio`: Default asyncio-based API.
- `pysnmp.hlapi.v1arch.asyncio.auth`: Authentication mechanisms.
- `pysnmp.hlapi.v1arch.asyncio.dispatch`: Dispatch mechanisms.
- `pysnmp.proto.rfc1902`: SNMP protocol objects.
- `pysnmp.proto.rfc1905`: EndOfMibView, NoSuchInstance, NoSuchObject.
- `pysnmp.smi.rfc1902`: SMI (Structure of Management Information) objects.
"""
from pysnmp.hlapi.v1arch.asyncio import *  # default is asyncio-based API
from pysnmp.hlapi.v1arch.asyncio.auth import *
from pysnmp.hlapi.v1arch.asyncio.dispatch import *
from pysnmp.proto.rfc1902 import *
from pysnmp.proto.rfc1905 import EndOfMibView
from pysnmp.proto.rfc1905 import NoSuchInstance
from pysnmp.proto.rfc1905 import NoSuchObject
from pysnmp.smi.rfc1902 import *



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp import error, nextid
from pysnmp.entity import config
from pysnmp.entity.engine import SnmpEngine
from pysnmp.hlapi.v3arch.asyncio.auth import CommunityData, UsmUserData
from pysnmp.hlapi.v3arch.asyncio.transport import AbstractTransportTarget

__all__ = ["CommandGeneratorLcdConfigurator", "NotificationOriginatorLcdConfigurator"]


class AbstractLcdConfigurator:
    next_id = nextid.Integer(0xFFFFFFFF)
    cache_keys = []

    def _get_cache(self, snmpEngine: SnmpEngine):
        cacheId = self.__class__.__name__
        cache = snmpEngine.get_user_context(cacheId)
        if cache is None:
            cache = {x: {} for x in self.cache_keys}
            snmpEngine.set_user_context(**{cacheId: cache})
        return cache

    def configure(self, snmpEngine, *args, **kwargs):
        pass

    def unconfigure(self, snmpEngine, *args, **kwargs):
        pass


class CommandGeneratorLcdConfigurator(AbstractLcdConfigurator):
    """Local configuration data (LCD) for Command Generator."""

    cache_keys = ["auth", "parm", "tran", "addr"]

    def configure(
        self,
        snmpEngine: SnmpEngine,
        authData,
        transportTarget: AbstractTransportTarget,
        contextName=b"",
        **options,
    ):
        """Configure command generator targets on the SNMP engine."""
        cache = self._get_cache(snmpEngine)
        if isinstance(authData, CommunityData):
            if authData.communityIndex not in cache["auth"]:
                config.add_v1_system(
                    snmpEngine,
                    authData.communityIndex,
                    authData.communityName,
                    authData.contextEngineId,
                    authData.context_name,
                    authData.tag,
                    authData.securityName,
                )
                cache["auth"][authData.communityIndex] = authData
        elif isinstance(authData, UsmUserData):
            authDataKey = authData.userName, authData.securityEngineId
            if authDataKey not in cache["auth"]:
                add_user = True

            elif self._usm_auth_changed(cache["auth"][authDataKey], authData):
                config.delete_v3_user(
                    snmpEngine, authData.userName, authData.securityEngineId
                )
                add_user = True

            else:
                add_user = False

            if add_user:
                config.add_v3_user(
                    snmpEngine,
                    authData.userName,
                    authData.authentication_protocol,
                    authData.authentication_key,
                    authData.privacy_protocol,
                    authData.privacy_key,
                    securityEngineId=authData.securityEngineId,
                    securityName=authData.securityName,
                    authKeyType=authData.authKeyType,
                    privKeyType=authData.privKeyType,
                )
                cache["auth"][authDataKey] = authData
        else:
            raise error.PySnmpError("Unsupported authentication object")

        paramsKey = (
            authData.securityName,
            authData.security_level,
            authData.message_processing_model,
        )
        if paramsKey in cache["parm"]:
            paramsName, useCount = cache["parm"][paramsKey]
            cache["parm"][paramsKey] = paramsName, useCount + 1
        else:
            paramsName = "p%s" % self.next_id()
            config.add_target_parameters(
                snmpEngine,
                paramsName,
                authData.securityName,
                authData.security_level,
                authData.message_processing_model,
            )
            cache["parm"][paramsKey] = paramsName, 1

        if transportTarget.TRANSPORT_DOMAIN in cache["tran"]:
            transport, useCount = cache["tran"][transportTarget.TRANSPORT_DOMAIN]
            transportTarget.verify_dispatcher_compatibility(snmpEngine)
            cache["tran"][transportTarget.TRANSPORT_DOMAIN] = transport, useCount + 1
        elif config.get_transport(snmpEngine, transportTarget.TRANSPORT_DOMAIN):
            transportTarget.verify_dispatcher_compatibility(snmpEngine)
        else:
            transport = transportTarget.open_client_mode()
            config.add_transport(
                snmpEngine, transportTarget.TRANSPORT_DOMAIN, transport
            )
            cache["tran"][transportTarget.TRANSPORT_DOMAIN] = transport, 1

        transportKey = (
            paramsName,
            transportTarget.TRANSPORT_DOMAIN,
            transportTarget.transport_address,
            transportTarget.timeout,
            transportTarget.retries,
            transportTarget.tagList,
            transportTarget.iface,
        )

        if transportKey in cache["addr"]:
            addrName, useCount = cache["addr"][transportKey]
            cache["addr"][transportKey] = addrName, useCount + 1
        else:
            addrName = "a%s" % self.next_id()
            config.add_target_address(
                snmpEngine,
                addrName,
                transportTarget.TRANSPORT_DOMAIN,
                transportTarget.transport_address,
                paramsName,
                transportTarget.timeout * 100,
                transportTarget.retries,
                transportTarget.tagList,
            )
            cache["addr"][transportKey] = addrName, 1

        return addrName, paramsName

    def unconfigure(
        self, snmpEngine: SnmpEngine, authData=None, contextName=b"", **options
    ):
        """Remove command generator targets from the SNMP engine."""
        cache = self._get_cache(snmpEngine)
        if authData:
            if isinstance(authData, CommunityData):
                authDataKey = authData.communityIndex
            elif isinstance(authData, UsmUserData):
                authDataKey = authData.userName, authData.securityEngineId
            else:
                raise error.PySnmpError("Unsupported authentication object")
            if authDataKey in cache["auth"]:
                authDataKeys = (authDataKey,)
            else:
                raise error.PySnmpError(f"Unknown authData {authData}")
        else:
            authDataKeys = list(cache["auth"].keys())

        addrNames, paramsNames = set(), set()

        for authDataKey in authDataKeys:
            authDataX = cache["auth"][authDataKey]
            del cache["auth"][authDataKey]
            if isinstance(authDataX, CommunityData):
                config.delete_v1_system(snmpEngine, authDataX.communityIndex)
            elif isinstance(authDataX, UsmUserData):
                config.delete_v3_user(
                    snmpEngine, authDataX.userName, authDataX.securityEngineId
                )
            else:
                raise error.PySnmpError("Unsupported authentication object")

            paramsKey = (
                authDataX.securityName,
                authDataX.security_level,
                authDataX.message_processing_model,
            )
            if paramsKey in cache["parm"]:
                paramsName, useCount = cache["parm"][paramsKey]
                useCount -= 1
                if useCount:
                    cache["parm"][paramsKey] = paramsName, useCount
                else:
                    del cache["parm"][paramsKey]
                    config.delete_target_parameters(snmpEngine, paramsName)
                    paramsNames.add(paramsName)
            else:
                raise error.PySnmpError(f"Unknown target {paramsKey}")

            addrKeys = [x for x in cache["addr"] if x[0] == paramsName]

            for addrKey in addrKeys:
                addrName, useCount = cache["addr"][addrKey]
                useCount -= 1
                if useCount:
                    cache["addr"][addrKey] = addrName, useCount
                else:
                    config.delete_target_address(snmpEngine, addrName)
                    del cache["addr"][addrKey]
                    addrNames.add(addrKey)

                    if addrKey[1] in cache["tran"]:
                        transport, useCount = cache["tran"][addrKey[1]]
                        if useCount > 1:
                            useCount -= 1
                            cache["tran"][addrKey[1]] = transport, useCount
                        else:
                            config.delete_transport(snmpEngine, addrKey[1])
                            transport.closeTransport()
                            del cache["tran"][addrKey[1]]

        return addrNames, paramsNames

    @staticmethod
    def _usm_auth_changed(cachedAuthData: UsmUserData, newAuthData: UsmUserData):
        changed = False

        changed |= cachedAuthData.authentication_key != newAuthData.authentication_key
        changed |= (
            cachedAuthData.authentication_protocol
            != newAuthData.authentication_protocol
        )
        changed |= cachedAuthData.privacy_key != newAuthData.privacy_key
        changed |= cachedAuthData.privacy_protocol != newAuthData.privacy_protocol

        return changed


class NotificationOriginatorLcdConfigurator(AbstractLcdConfigurator):
    """Local configuration data (LCD) for Notification Originator."""

    cache_keys = ["auth", "name"]
    _lcd_configurator = CommandGeneratorLcdConfigurator()

    def configure(
        self,
        snmpEngine: SnmpEngine,
        authData: "CommunityData | UsmUserData",
        transportTarget: AbstractTransportTarget,
        notifyType: str,
        contextName=None,
        **options,
    ):
        """Configure notification targets on the SNMP engine."""
        cache = self._get_cache(snmpEngine)
        notifyName = None

        # Create matching transport tags if not given by user. Not good!
        if not transportTarget.tagList:
            transportTarget.tagList = str(
                hash((authData.securityName, transportTarget.transport_address))
            )
        if isinstance(authData, CommunityData) and not authData.tag:
            authData.tag = transportTarget.tagList.split()[0]

        addrName, paramsName = self._lcd_configurator.configure(
            snmpEngine, authData, transportTarget, contextName, **options
        )
        tagList = transportTarget.tagList.split()
        if not tagList:
            tagList = [""]
        for tag in tagList:
            notifyNameKey = paramsName, tag, notifyType
            if notifyNameKey in cache["name"]:
                notifyName, paramsName, useCount = cache["name"][notifyNameKey]
                cache["name"][notifyNameKey] = notifyName, paramsName, useCount + 1
            else:
                notifyName = "n%s" % self.next_id()
                config.add_notification_target(
                    snmpEngine, notifyName, paramsName, tag, notifyType
                )
                cache["name"][notifyNameKey] = notifyName, paramsName, 1
        authDataKey = (
            authData.securityName,
            authData.security_model,
            authData.security_level,
            contextName,
        )
        if authDataKey in cache["auth"]:
            authDataX, subTree, useCount = cache["auth"][authDataKey]
            cache["auth"][authDataKey] = authDataX, subTree, useCount + 1
        else:
            subTree = (1, 3, 6)
            config.add_vacm_user(
                snmpEngine,
                authData.security_model,
                authData.securityName,
                authData.security_level,
                (),
                (),
                subTree,
                contextName=contextName,
            )
            cache["auth"][authDataKey] = authData, subTree, 1

        return notifyName

    def unconfigure(
        self, snmpEngine: SnmpEngine, authData=None, contextName=b"", **options
    ):
        """Remove notification targets from the SNMP engine."""
        cache = self._get_cache(snmpEngine)
        if authData:
            authDataKey = (
                authData.securityName,
                authData.securityModel,
                authData.securityLevel,
                contextName,
            )
            if authDataKey in cache["auth"]:
                authDataKeys = (authDataKey,)
            else:
                raise error.PySnmpError(f"Unknown authData {authData}")
        else:
            authDataKeys = tuple(cache["auth"])

        addrNames, paramsNames = self._lcd_configurator.unconfigure(
            snmpEngine, authData, contextName, **options
        )

        notifyAndParamsNames = [
            (cache["name"][x], x) for x in cache["name"].keys() if x[0] in paramsNames
        ]

        for (notifyName, paramsName, useCount), notifyNameKey in notifyAndParamsNames:
            useCount -= 1
            if useCount:
                cache["name"][notifyNameKey] = notifyName, paramsName, useCount
            else:
                config.delete_notification_target(snmpEngine, notifyName, paramsName)
                del cache["name"][notifyNameKey]

        for authDataKey in authDataKeys:
            authDataX, subTree, useCount = cache["auth"][authDataKey]
            useCount -= 1
            if useCount:
                cache["auth"][authDataKey] = authDataX, subTree, useCount
            else:
                config.delTrapUser(
                    snmpEngine,
                    authDataX.securityModel,
                    authDataX.securityName,
                    authDataX.securityLevel,
                    subTree,
                )
                del cache["auth"][authDataKey]



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from pyasn1.codec.ber import decoder, eoo
from pyasn1.type import constraint, namedtype, univ
from pysnmp import debug
from pysnmp.proto import api, errind, error, rfc1905, rfc3411
from pysnmp.proto.mpmod.base import AbstractMessageProcessingModel
from pysnmp.proto.secmod.base import AbstractSecurityModel

if TYPE_CHECKING:
    from pysnmp.entity.engine import SnmpEngine


# API to rfc1905 protocol objects
pMod = api.PROTOCOL_MODULES[api.SNMP_VERSION_2C]  # noqa: N816


# SNMPv3 message format


class ScopedPDU(univ.Sequence):
    """Scoped PDU syntax."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("contextEngineId", univ.OctetString()),
        namedtype.NamedType("contextName", univ.OctetString()),
        namedtype.NamedType("data", rfc1905.PDUs()),
    )


class ScopedPduData(univ.Choice):
    """Scoped PDU data syntax."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("plaintext", ScopedPDU()),
        namedtype.NamedType("encryptedPDU", univ.OctetString()),
    )


class HeaderData(univ.Sequence):
    """SNMPv3 message header data syntax."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType(
            "msgID",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(0, 2147483647)
            ),
        ),
        namedtype.NamedType(
            "msgMaxSize",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(484, 2147483647)
            ),
        ),
        namedtype.NamedType(
            "msgFlags",
            univ.OctetString().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(1, 1)
            ),
        ),
        # NOTE (etingof): constrain SNMPv3 message to only USM+ security models
        # because SNMPv1/v2c seems incompatible in pysnmp implementation, not sure
        # if it's intended by the SNMP standard at all...
        namedtype.NamedType(
            "msgSecurityModel",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(3, 2147483647)
            ),
        ),
    )


class SNMPv3Message(univ.Sequence):
    """SNMPv3 message syntax."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType(
            "msgVersion",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(0, 2147483647)
            ),
        ),
        namedtype.NamedType("msgGlobalData", HeaderData()),
        namedtype.NamedType("msgSecurityParameters", univ.OctetString()),
        namedtype.NamedType("msgData", ScopedPduData()),
    )


# XXX move somewhere?
_snmpErrors = {  # noqa: N816
    (1, 3, 6, 1, 6, 3, 15, 1, 1, 1, 0): errind.unsupportedSecurityLevel,
    (1, 3, 6, 1, 6, 3, 15, 1, 1, 2, 0): errind.notInTimeWindow,
    (1, 3, 6, 1, 6, 3, 15, 1, 1, 3, 0): errind.unknownUserName,
    (1, 3, 6, 1, 6, 3, 15, 1, 1, 4, 0): errind.unknownEngineID,
    (1, 3, 6, 1, 6, 3, 15, 1, 1, 5, 0): errind.wrongDigest,
    (1, 3, 6, 1, 6, 3, 15, 1, 1, 6, 0): errind.decryptionError,
}


class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
    """Implement SNMPv3 message processing model."""

    MESSAGE_PROCESSING_MODEL_ID = univ.Integer(3)  # SNMPv3
    SNMP_MSG_SPEC = SNMPv3Message
    _empty_string = univ.OctetString("")
    _message_flags = {
        0: univ.OctetString("\x00"),
        1: univ.OctetString("\x01"),
        3: univ.OctetString("\x03"),
        4: univ.OctetString("\x04"),
        5: univ.OctetString("\x05"),
        7: univ.OctetString("\x07"),
    }

    def __init__(self):
        """Create SNMPv3 message processing model instance."""
        AbstractMessageProcessingModel.__init__(self)
        self.__scopedPDU = ScopedPDU()
        self.__engineIdCache = {}
        self.__engineIdCacheExpQueue = {}
        self.__expirationTimer = 0

    def get_peer_engine_info(self, transportDomain, transportAddress):
        """Return cached peer SNMP engine data."""
        k = transportDomain, transportAddress
        if k in self.__engineIdCache:
            return (
                self.__engineIdCache[k]["securityEngineId"],
                self.__engineIdCache[k]["contextEngineId"],
                self.__engineIdCache[k]["contextName"],
            )
        else:
            return None, None, None

    # 7.1.1a
    def prepare_outgoing_message(
        self,
        snmpEngine: SnmpEngine,
        transportDomain,
        transportAddress,
        messageProcessingModel,
        securityModel,
        securityName,
        securityLevel,
        contextEngineId,
        contextName,
        pduVersion,
        pdu,
        expectResponse,
        sendPduHandle,
    ):
        """Prepare SNMP message for dispatch."""
        (snmpEngineID,) = snmpEngine.get_mib_builder().import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        snmpEngineID = snmpEngineID.syntax

        # 7.1.1b
        msgID = self._cache.new_message_id()

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareOutgoingMessage: new msgID %s" % msgID
        )

        k = (transportDomain, transportAddress)
        if k in self.__engineIdCache:
            peerSnmpEngineData = self.__engineIdCache[k]
        else:
            peerSnmpEngineData = None

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareOutgoingMessage: peer SNMP engine data {} for transport {}, address {}".format(
                peerSnmpEngineData, transportDomain, transportAddress
            )
        )

        # 7.1.4
        if contextEngineId is None:
            if peerSnmpEngineData is None:
                contextEngineId = snmpEngineID
            else:
                contextEngineId = peerSnmpEngineData["contextEngineId"]
                # Defaulting contextEngineID to securityEngineId should
                # probably be done on Agent side (see 7.1.3.d.2,) so this
                # is a sort of workaround.
                if not contextEngineId:
                    contextEngineId = peerSnmpEngineData["securityEngineId"]
        # 7.1.5
        if not contextName:
            contextName = self._empty_string

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareOutgoingMessage: using contextEngineId {contextEngineId!r}, contextName {contextName!r}"
        )

        # 7.1.6
        scopedPDU = self.__scopedPDU
        scopedPDU.setComponentByPosition(0, contextEngineId)
        scopedPDU.setComponentByPosition(1, contextName)
        scopedPDU.setComponentByPosition(2)
        scopedPDU.getComponentByPosition(2).setComponentByType(
            pdu.tagSet,
            pdu,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 7.1.7
        msg = self._snmpMsgSpec

        # 7.1.7a
        msg.setComponentByPosition(
            0,
            self.MESSAGE_PROCESSING_MODEL_ID,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )
        headerData = msg.setComponentByPosition(1).getComponentByPosition(1)

        # 7.1.7b
        headerData.setComponentByPosition(
            0, msgID, verifyConstraints=False, matchTags=False, matchConstraints=False
        )

        (snmpEngineMaxMessageSize,) = snmpEngine.get_mib_builder().import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineMaxMessageSize"
        )

        # 7.1.7c
        # XXX need to coerce MIB value as it has incompatible constraints set
        headerData.setComponentByPosition(
            1,
            snmpEngineMaxMessageSize.syntax,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 7.1.7d
        msgFlags = 0
        if securityLevel == 1:
            pass
        elif securityLevel == 2:
            msgFlags |= 0x01
        elif securityLevel == 3:
            msgFlags |= 0x03
        else:
            raise error.ProtocolError("Unknown securityLevel %s" % securityLevel)

        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:
            msgFlags |= 0x04

        headerData.setComponentByPosition(
            2,
            self._message_flags[msgFlags],
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 7.1.7e
        # XXX need to coerce MIB value as it has incompatible constraints set
        headerData.setComponentByPosition(3, int(securityModel))

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareOutgoingMessage: {msg.prettyPrint()}"
        )

        if securityModel in snmpEngine.security_models:
            smHandler = snmpEngine.security_models[securityModel]
        else:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedSecurityModel
            )

        # 7.1.9.a
        if pdu.tagSet in rfc3411.UNCONFIRMED_CLASS_PDUS:
            securityEngineId = snmpEngineID

        else:
            if peerSnmpEngineData is None:
                debug.logger & debug.FLAG_MP and debug.logger(
                    "prepareOutgoingMessage: peer SNMP engine is not known"
                )

                securityEngineId = None

            else:
                securityEngineId = peerSnmpEngineData["securityEngineId"]

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareOutgoingMessage: securityModel {!r}, securityEngineId {!r}, securityName {!r}, securityLevel {!r}".format(
                securityModel, securityEngineId, securityName, securityLevel
            )
        )

        # 7.1.9.b
        (securityParameters, wholeMsg) = smHandler.generate_request_message(
            snmpEngine,
            self.MESSAGE_PROCESSING_MODEL_ID,
            msg,
            snmpEngineMaxMessageSize.syntax,
            securityModel,
            securityEngineId,
            securityName,
            securityLevel,
            scopedPDU,
        )

        # Message size constraint verification
        if len(wholeMsg) > snmpEngineMaxMessageSize.syntax:
            raise error.StatusInformation(errorIndication=errind.tooBig)

        # 7.1.9.c
        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:
            # XXX rfc bug? why stateReference should be created?
            self._cache.push_by_message_id(
                msgID,
                sendPduHandle=sendPduHandle,
                msgID=msgID,
                snmpEngineID=snmpEngineID,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                transportDomain=transportDomain,
                transportAddress=transportAddress,
            )

        snmpEngine.observer.store_execution_context(
            snmpEngine,
            "rfc3412.prepareOutgoingMessage",
            dict(
                transportDomain=transportDomain,
                transportAddress=transportAddress,
                wholeMsg=wholeMsg,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                pdu=pdu,
            ),
        )
        snmpEngine.observer.clear_execution_context(
            snmpEngine, "rfc3412.prepareOutgoingMessage"
        )

        return transportDomain, transportAddress, wholeMsg

    def prepare_response_message(
        self,
        snmpEngine: SnmpEngine,
        messageProcessingModel,
        securityModel,
        securityName,
        securityLevel,
        contextEngineId,
        contextName,
        pduVersion,
        pdu,
        maxSizeResponseScopedPDU,
        stateReference,
        statusInformation,
    ):
        """Prepare SNMP message for response."""
        (snmpEngineID,) = snmpEngine.get_mib_builder().import_symbols(  # type: ignore
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        snmpEngineID = snmpEngineID.syntax

        # 7.1.2.b
        cachedParams = self._cache.pop_by_state_reference(stateReference)
        msgID = cachedParams["msgID"]
        contextEngineId = cachedParams["contextEngineId"]
        contextName = cachedParams["contextName"]
        securityModel = cachedParams["securityModel"]
        securityName = cachedParams["securityName"]
        securityLevel = cachedParams["securityLevel"]
        securityStateReference = cachedParams["securityStateReference"]
        reportableFlag = cachedParams["reportableFlag"]
        maxMessageSize = cachedParams["msgMaxSize"]
        transportDomain = cachedParams["transportDomain"]
        transportAddress = cachedParams["transportAddress"]

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareResponseMessage: stateReference %s" % stateReference
        )

        # 7.1.3
        if statusInformation is not None and "oid" in statusInformation:
            # 7.1.3a
            if pdu is None:
                pduType = None
            else:
                requestID = pdu.getComponentByPosition(0)
                pduType = pdu.tagSet

            # 7.1.3b
            if (
                pdu is None
                and not reportableFlag
                or pduType is not None
                and pduType not in rfc3411.CONFIRMED_CLASS_PDUS
            ):
                if securityModel in snmpEngine.security_models:
                    smHandler = snmpEngine.security_models[securityModel]
                    smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.loopTerminated)

            # 7.1.3c
            reportPDU = rfc1905.ReportPDU()
            pMod.apiPDU.set_varbinds(
                reportPDU, ((statusInformation["oid"], statusInformation["val"]),)
            )
            pMod.apiPDU.set_error_status(reportPDU, 0)
            pMod.apiPDU.set_error_index(reportPDU, 0)
            if pdu is None:
                pMod.apiPDU.set_request_id(reportPDU, 0)
            else:
                # noinspection PyUnboundLocalVariable
                pMod.apiPDU.set_request_id(reportPDU, requestID)

            # 7.1.3d.1
            if "securityLevel" in statusInformation:
                securityLevel = statusInformation["securityLevel"]
            else:
                securityLevel = 1

            # 7.1.3d.2
            if "contextEngineId" in statusInformation:
                contextEngineId = statusInformation["contextEngineId"]
            else:
                contextEngineId = snmpEngineID

            # 7.1.3d.3
            if "contextName" in statusInformation:
                contextName = statusInformation["contextName"]
            else:
                contextName = ""

            # 7.1.3e
            pdu = reportPDU

            debug.logger & debug.FLAG_MP and debug.logger(
                "prepareResponseMessage: prepare report PDU for statusInformation %s"
                % statusInformation
            )
        # 7.1.4
        if not contextEngineId:
            contextEngineId = snmpEngineID  # XXX impl-dep manner

        # 7.1.5
        if not contextName:
            contextName = self._empty_string

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareResponseMessage: using contextEngineId {contextEngineId!r}, contextName {contextName!r}"
        )

        # 7.1.6
        scopedPDU = self.__scopedPDU
        scopedPDU.setComponentByPosition(0, contextEngineId)
        scopedPDU.setComponentByPosition(1, contextName)
        scopedPDU.setComponentByPosition(2)
        scopedPDU.getComponentByPosition(2).setComponentByType(
            pdu.tagSet,
            pdu,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 7.1.7
        msg = self._snmpMsgSpec

        # 7.1.7a
        msg.setComponentByPosition(
            0,
            self.MESSAGE_PROCESSING_MODEL_ID,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        headerData = msg.setComponentByPosition(1).getComponentByPosition(1)

        # 7.1.7b
        headerData.setComponentByPosition(
            0, msgID, verifyConstraints=False, matchTags=False, matchConstraints=False
        )

        (snmpEngineMaxMessageSize,) = snmpEngine.get_mib_builder().import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineMaxMessageSize"
        )

        # 7.1.7c
        # XXX need to coerce MIB value as it has incompatible constraints set
        headerData.setComponentByPosition(
            1,
            snmpEngineMaxMessageSize.syntax,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 7.1.7d
        msgFlags = 0
        if securityLevel == 1:
            pass
        elif securityLevel == 2:
            msgFlags |= 0x01
        elif securityLevel == 3:
            msgFlags |= 0x03
        else:
            raise error.ProtocolError("Unknown securityLevel %s" % securityLevel)

        if pdu.tagSet in rfc3411.CONFIRMED_CLASS_PDUS:  # XXX not needed?
            msgFlags |= 0x04

        headerData.setComponentByPosition(
            2,
            self._message_flags[msgFlags],
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        # 7.1.7e
        headerData.setComponentByPosition(
            3,
            securityModel,
            verifyConstraints=False,
            matchTags=False,
            matchConstraints=False,
        )

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareResponseMessage: {msg.prettyPrint()}"
        )

        if securityModel in snmpEngine.security_models:
            smHandler = snmpEngine.security_models[securityModel]
        else:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedSecurityModel
            )

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareResponseMessage: securityModel {!r}, securityEngineId {!r}, securityName {!r}, securityLevel {!r}".format(
                securityModel, snmpEngineID, securityName, securityLevel
            )
        )

        # 7.1.8a
        try:
            (securityParameters, wholeMsg) = smHandler.generate_response_message(
                snmpEngine,
                self.MESSAGE_PROCESSING_MODEL_ID,
                msg,
                snmpEngineMaxMessageSize.syntax,
                securityModel,
                snmpEngineID,
                securityName,
                securityLevel,
                scopedPDU,
                securityStateReference,
                statusInformation.get(
                    "errorIndication"
                ),  # Use get() to handle missing errorIndication
            )  # type: ignore
        except error.StatusInformation:
            # 7.1.8.b
            raise

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareResponseMessage: SM finished"
        )

        # Message size constraint verification
        if len(wholeMsg) > min(snmpEngineMaxMessageSize.syntax, maxMessageSize):
            raise error.StatusInformation(errorIndication=errind.tooBig)

        snmpEngine.observer.store_execution_context(
            snmpEngine,
            "rfc3412.prepareResponseMessage",
            dict(
                transportDomain=transportDomain,
                transportAddress=transportAddress,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                contextEngineId=contextEngineId,
                contextName=contextName,
                securityEngineId=snmpEngineID,
                pdu=pdu,
            ),
        )
        snmpEngine.observer.clear_execution_context(
            snmpEngine, "rfc3412.prepareResponseMessage"
        )

        return transportDomain, transportAddress, wholeMsg

    # 7.2.1

    def prepare_data_elements(
        self, snmpEngine: SnmpEngine, transportDomain, transportAddress, wholeMsg
    ):
        """Prepare SNMP message data elements."""
        # 7.2.2
        msg, restOfwholeMsg = decoder.decode(wholeMsg, asn1Spec=self._snmpMsgSpec)

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareDataElements: {msg.prettyPrint()}"
        )

        if eoo.endOfOctets.isSameTypeWith(msg):
            raise error.StatusInformation(errorIndication=errind.parseError)

        # 7.2.3
        headerData = msg.getComponentByPosition(1)
        msgVersion = messageProcessingModel = msg.getComponentByPosition(0)
        msgID = headerData.getComponentByPosition(0)
        (msgFlags,) = headerData.getComponentByPosition(2).asNumbers()
        maxMessageSize = headerData.getComponentByPosition(1)
        securityModel = headerData.getComponentByPosition(3)
        securityParameters = msg.getComponentByPosition(2)

        debug.logger & debug.FLAG_MP and debug.logger(
            "prepareDataElements: msg data msgVersion {} msgID {} securityModel {}".format(
                msgVersion, msgID, securityModel
            )
        )

        # 7.2.4
        if securityModel not in snmpEngine.security_models:
            (snmpUnknownSecurityModels,) = snmpEngine.get_mib_builder().import_symbols(
                "__SNMP-MPD-MIB", "snmpUnknownSecurityModels"
            )
            snmpUnknownSecurityModels.syntax += 1
            raise error.StatusInformation(
                errorIndication=errind.unsupportedSecurityModel
            )

        # 7.2.5
        if msgFlags & 0x03 == 0x00:
            securityLevel = 1  # noAuthNoPriv
        elif (msgFlags & 0x03) == 0x01:
            securityLevel = 2  # authNoPriv
        elif (msgFlags & 0x03) == 0x03:
            securityLevel = 3  # authPriv
        else:
            (snmpInvalidMsgs,) = snmpEngine.get_mib_builder().import_symbols(
                "__SNMP-MPD-MIB", "snmpInvalidMsgs"
            )
            snmpInvalidMsgs.syntax += 1
            raise error.StatusInformation(errorIndication=errind.invalidMsg)

        if msgFlags & 0x04:
            reportableFlag = 1
        else:
            reportableFlag = 0

        # 7.2.6
        smHandler: AbstractSecurityModel = snmpEngine.security_models[securityModel]
        try:
            (
                securityEngineId,
                securityName,
                scopedPDU,
                maxSizeResponseScopedPDU,
                securityStateReference,
            ) = smHandler.process_incoming_message(
                snmpEngine,
                messageProcessingModel,
                maxMessageSize,
                securityParameters,
                securityModel,
                securityLevel,
                wholeMsg,
                msg,
            )  # type: ignore
            debug.logger & debug.FLAG_MP and debug.logger(
                "prepareDataElements: SM succeeded"
            )

        except error.StatusInformation:
            statusInformation, origTraceback = sys.exc_info()[1:3]

            debug.logger & debug.FLAG_MP and debug.logger(
                "prepareDataElements: SM failed, statusInformation %s"
                % statusInformation
            )

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc3412.prepareDataElements:sm-failure",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityLevel=securityLevel,
                    securityParameters=securityParameters,
                    statusInformation=statusInformation,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc3412.prepareDataElements:sm-failure"
            )

            if "errorIndication" in statusInformation:
                # 7.2.6a
                if "oid" in statusInformation:
                    # 7.2.6a1
                    securityStateReference = statusInformation["securityStateReference"]
                    contextEngineId = statusInformation["contextEngineId"]
                    contextName = statusInformation["contextName"]
                    if "scopedPDU" in statusInformation:
                        scopedPDU = statusInformation["scopedPDU"]
                        pdu = scopedPDU.getComponentByPosition(2).getComponent()
                    else:
                        pdu = None
                    maxSizeResponseScopedPDU = statusInformation[
                        "maxSizeResponseScopedPDU"
                    ]
                    securityName = None  # XXX secmod cache used

                    # 7.2.6a2
                    stateReference = self._cache.new_state_reference()
                    self._cache.push_by_state_reference(
                        stateReference,
                        msgVersion=messageProcessingModel,
                        msgID=msgID,
                        contextEngineId=contextEngineId,
                        contextName=contextName,
                        securityModel=securityModel,
                        securityName=securityName,
                        securityLevel=securityLevel,
                        securityStateReference=securityStateReference,
                        reportableFlag=reportableFlag,
                        msgMaxSize=maxMessageSize,
                        maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                        transportDomain=transportDomain,
                        transportAddress=transportAddress,
                    )

                    # 7.2.6a3
                    try:
                        snmpEngine.message_dispatcher.return_response_pdu(
                            snmpEngine,
                            3,
                            securityModel,
                            securityName,
                            securityLevel,
                            contextEngineId,
                            contextName,
                            1,
                            pdu,
                            maxSizeResponseScopedPDU,
                            stateReference,
                            statusInformation,
                        )
                    except error.StatusInformation:
                        pass

                    debug.logger & debug.FLAG_MP and debug.logger(
                        "prepareDataElements: error reported"
                    )

            # 7.2.6b
            try:
                raise statusInformation.with_traceback(origTraceback)
            finally:
                # Break cycle between locals and traceback object
                # (seems to be irrelevant on Py3 but just in case)
                del origTraceback
        else:
            # Sniff for engineIdCache
            k = (transportDomain, transportAddress)
            if k not in self.__engineIdCache:
                contextEngineId = scopedPDU[0]
                contextName = scopedPDU[1]
                pdus = scopedPDU[2]
                pdu = pdus.getComponent()
                # Here we assume that authentic/default EngineIDs
                # come only in the course of engine-to-engine communication.
                if pdu.tagSet in rfc3411.INTERNAL_CLASS_PDUS:
                    self.__engineIdCache[k] = {
                        "securityEngineId": securityEngineId,
                        "contextEngineId": contextEngineId,
                        "contextName": contextName,
                    }

                    expireAt = int(
                        self.__expirationTimer
                        + 300 / snmpEngine.transport_dispatcher.get_timer_resolution()
                    )
                    if expireAt not in self.__engineIdCacheExpQueue:
                        self.__engineIdCacheExpQueue[expireAt] = []
                    self.__engineIdCacheExpQueue[expireAt].append(k)

                    debug.logger & debug.FLAG_MP and debug.logger(
                        "prepareDataElements: cache securityEngineId {!r} for {!r} {!r}".format(
                            securityEngineId, transportDomain, transportAddress
                        )
                    )

        (snmpEngineID,) = snmpEngine.get_mib_builder().import_symbols(
            "__SNMP-FRAMEWORK-MIB", "snmpEngineID"
        )
        snmpEngineID = snmpEngineID.syntax

        # 7.2.7 XXX PDU would be parsed here?
        contextEngineId = scopedPDU[0]
        contextName = scopedPDU[1]
        pdu = scopedPDU[2]

        pdu = pdu.getComponent()  # PDUs

        # 7.2.8
        pduVersion = api.SNMP_VERSION_2C

        # 7.2.9
        pduType = pdu.tagSet

        # 7.2.10
        if (
            pduType in rfc3411.RESPONSE_CLASS_PDUS
            or pduType in rfc3411.INTERNAL_CLASS_PDUS
        ):
            # 7.2.10a
            try:
                cachedReqParams = self._cache.pop_by_message_id(msgID)
            except error.ProtocolError:
                smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.dataMismatch)
            # 7.2.10b
            sendPduHandle = cachedReqParams["sendPduHandle"]
        else:
            sendPduHandle = None

        debug.logger & debug.FLAG_MP and debug.logger(
            f"prepareDataElements: using sendPduHandle {sendPduHandle} for msgID {msgID}"
        )

        # 7.2.11
        if pduType in rfc3411.INTERNAL_CLASS_PDUS:
            # 7.2.11a
            varBinds = pMod.apiPDU.get_varbinds(pdu)
            if varBinds:
                statusInformation = error.StatusInformation(
                    errorIndication=_snmpErrors.get(
                        varBinds[0][0],
                        errind.ReportPduReceived(varBinds[0][0].prettyPrint()),
                    ),
                    oid=varBinds[0][0],
                    val=varBinds[0][1],
                    sendPduHandle=sendPduHandle,
                )
            else:
                statusInformation = error.StatusInformation(sendPduHandle=sendPduHandle)

            # 7.2.11b (incomplete implementation)

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc3412.prepareDataElements:internal",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc3412.prepareDataElements:internal"
            )

            # 7.2.11c
            smHandler.release_state_information(securityStateReference)

            # 7.2.11d
            # no-op

            # 7.2.11e XXX may need to pass Reports up to app in some cases...
            raise statusInformation

        statusInformation = None  # no errors ahead

        # 7.2.12
        if pduType in rfc3411.RESPONSE_CLASS_PDUS:
            # 7.2.12a -> no-op

            # 7.2.12b
            # noinspection PyUnboundLocalVariable
            if (
                securityModel != cachedReqParams["securityModel"]
                or securityName != cachedReqParams["securityName"]
                or securityLevel != cachedReqParams["securityLevel"]
                or contextEngineId != cachedReqParams["contextEngineId"]
                or contextName != cachedReqParams["contextName"]
            ):
                smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.dataMismatch)

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc3412.prepareDataElements:response",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc3412.prepareDataElements:response"
            )

            # 7.2.12c
            smHandler.release_state_information(securityStateReference)
            stateReference = None

            # 7.2.12d
            return (
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                pdu,
                pduType,
                sendPduHandle,
                maxSizeResponseScopedPDU,
                statusInformation,
                stateReference,
            )

        # 7.2.13
        if pduType in rfc3411.CONFIRMED_CLASS_PDUS:
            # 7.2.13a
            if securityEngineId != snmpEngineID:
                smHandler.release_state_information(securityStateReference)
                raise error.StatusInformation(errorIndication=errind.engineIDMismatch)

            # 7.2.13b
            stateReference = self._cache.new_state_reference()
            self._cache.push_by_state_reference(
                stateReference,
                msgVersion=messageProcessingModel,
                msgID=msgID,
                contextEngineId=contextEngineId,
                contextName=contextName,
                securityModel=securityModel,
                securityName=securityName,
                securityLevel=securityLevel,
                securityStateReference=securityStateReference,
                reportableFlag=reportableFlag,
                msgMaxSize=maxMessageSize,
                maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
                transportDomain=transportDomain,
                transportAddress=transportAddress,
            )

            debug.logger & debug.FLAG_MP and debug.logger(
                "prepareDataElements: new stateReference %s" % stateReference
            )

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc3412.prepareDataElements:confirmed",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc3412.prepareDataElements:confirmed"
            )

            # 7.2.13c
            return (
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                pdu,
                pduType,
                sendPduHandle,
                maxSizeResponseScopedPDU,
                statusInformation,
                stateReference,
            )

        # 7.2.14
        if pduType in rfc3411.UNCONFIRMED_CLASS_PDUS:
            # Pass new stateReference to let app browse request details
            stateReference = self._cache.new_state_reference()

            snmpEngine.observer.store_execution_context(
                snmpEngine,
                "rfc3412.prepareDataElements:unconfirmed",
                dict(
                    transportDomain=transportDomain,
                    transportAddress=transportAddress,
                    securityModel=securityModel,
                    securityName=securityName,
                    securityLevel=securityLevel,
                    contextEngineId=contextEngineId,
                    contextName=contextName,
                    securityEngineId=securityEngineId,
                    pdu=pdu,
                ),
            )
            snmpEngine.observer.clear_execution_context(
                snmpEngine, "rfc3412.prepareDataElements:unconfirmed"
            )

            # This is not specified explicitly in RFC
            smHandler.release_state_information(securityStateReference)

            return (
                messageProcessingModel,
                securityModel,
                securityName,
                securityLevel,
                contextEngineId,
                contextName,
                pduVersion,
                pdu,
                pduType,
                sendPduHandle,
                maxSizeResponseScopedPDU,
                statusInformation,
                stateReference,
            )

        smHandler.release_state_information(securityStateReference)
        raise error.StatusInformation(errorIndication=errind.unsupportedPDUtype)

    def __expire_engines_info(self):
        if self.__expirationTimer in self.__engineIdCacheExpQueue:
            for engineKey in self.__engineIdCacheExpQueue[self.__expirationTimer]:
                del self.__engineIdCache[engineKey]
                debug.logger & debug.FLAG_MP and debug.logger(
                    f"__expireEnginesInfo: expiring {engineKey!r}"
                )
            del self.__engineIdCacheExpQueue[self.__expirationTimer]
        self.__expirationTimer += 1

    def receive_timer_tick(self, snmpEngine: SnmpEngine, timeNow):
        """Process periodic timer tick."""
        self.__expire_engines_info()
        AbstractMessageProcessingModel.receive_timer_tick(self, snmpEngine, timeNow)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from typing import Tuple


from pysnmp import error
from pysnmp.carrier.base import AbstractTransport
from pysnmp.entity.engine import SnmpEngine

__all__ = []


class AbstractTransportTarget:
    retries: int
    timeout: float
    transport: "AbstractTransport | None"
    transport_address: Tuple[str, int]

    TRANSPORT_DOMAIN = None
    PROTO_TRANSPORT = AbstractTransport

    def __init__(
        self,
        timeout: float = 1,
        retries: int = 5,
        tagList=b"",
    ):
        self.timeout = timeout
        self.retries = retries
        self.tagList = tagList
        self.iface = None
        self.transport = None

        if not hasattr(self, "transport_address"):
            raise Exception(
                f"Please call .create() to construct {self.__class__.__name__} object"
            )

    @classmethod
    async def create(cls, address: Tuple[str, int], *args, **kwargs):
        """
        Asynchronously creates an instance of the class with the given address.

        Args:
            cls (Type): The class to create an instance of.
            address (Tuple[str, int]): A tuple containing the address as a string and the port as an integer.
            *args: Additional positional arguments to pass to the class initializer.
            **kwargs: Additional keyword arguments to pass to the class initializer.

        Returns:
            An instance of the class with the resolved transport address.

        """
        self = cls.__new__(cls)
        transportAddr = address
        self.transport_address = await self._resolve_address(transportAddr)
        self.__init__(*args, **kwargs)
        return self

    def __repr__(self):
        return "{}({!r}, timeout={!r}, retries={!r}, tagList={!r})".format(
            self.__class__.__name__,
            self.transport_address,
            self.timeout,
            self.retries,
            self.tagList,
        )

    def get_transport_info(self):
        return self.TRANSPORT_DOMAIN, self.transport_address

    def set_local_address(self, iface):
        """Set source address.

        Parameters
        ----------
        iface : tuple
            Indicates network address of a local interface from which SNMP packets will be originated.
            Format is the same as of `transportAddress`.

        Returns
        -------
            self

        """
        self.iface = iface
        return self

    def open_client_mode(self):
        self.transport = self.PROTO_TRANSPORT().open_client_mode(self.iface)
        return self.transport

    def verify_dispatcher_compatibility(self, snmpEngine: SnmpEngine):
        if (
            snmpEngine.transport_dispatcher is None
            or not self.PROTO_TRANSPORT.is_compatible_with_dispatcher(
                snmpEngine.transport_dispatcher
            )
        ):
            raise error.PySnmpError(
                "Transport {!r} is not compatible with dispatcher {!r}".format(
                    self.PROTO_TRANSPORT, snmpEngine.transport_dispatcher
                )
            )

    async def _resolve_address(self, transportAddr: Tuple) -> Tuple[str, int]:
        raise NotImplementedError()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import marshal
import os
import struct
import sys
import time
import traceback
from typing import Any
import warnings
from errno import ENOENT
from importlib.machinery import BYTECODE_SUFFIXES, SOURCE_SUFFIXES
from importlib.util import MAGIC_NUMBER as PY_MAGIC_NUMBER

from pysnmp import debug, version as pysnmp_version
from pysnmp.smi import error

if __debug__:
    import runpy


PY_SUFFIXES = SOURCE_SUFFIXES + BYTECODE_SUFFIXES

classTypes = (type,)  # noqa: N816


class __AbstractMibSource:
    def __init__(self, srcName):
        self._srcName = srcName
        self.__inited = None
        debug.logger & debug.FLAG_BLD and debug.logger("trying %s" % self)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._srcName!r})"

    def _unique_names(self, files):
        u = set()

        for f in files:
            if f.startswith("__init__."):
                continue

            u.update(f[: -len(sfx)] for sfx in PY_SUFFIXES if f.endswith(sfx))

        return tuple(u)

    # MibSource API follows

    def full_path(self, f="", sfx=""):
        return self._srcName + (f and (os.sep + f + sfx) or "")

    def init(self):
        if self.__inited is None:
            self.__inited = self._init()
            if self.__inited is self:
                self.__inited = True
        if self.__inited is True:
            return self

        else:
            return self.__inited

    def listdir(self):
        return self._listdir()

    def read(self, f):
        pycTime = pyTime = -1

        for pycSfx in BYTECODE_SUFFIXES:
            try:
                pycData, pycPath = self._get_data(f + pycSfx, "rb")

            except OSError:
                why = sys.exc_info()[1]
                if ENOENT == -1 or why.errno == ENOENT:
                    debug.logger & debug.FLAG_BLD and debug.logger(
                        f"file {f + pycSfx} access error: {why}"
                    )

                else:
                    raise error.MibLoadError(
                        f"MIB file {f + pycSfx} access error: {why}"
                    )

            else:
                if PY_MAGIC_NUMBER == pycData[:4]:
                    pycData = pycData[4:]
                    pycTime = struct.unpack("<L", pycData[:4])[0]
                    pycData = pycData[4:]
                    debug.logger & debug.FLAG_BLD and debug.logger(
                        "file %s mtime %d" % (pycPath, pycTime)
                    )
                    break

                else:
                    debug.logger & debug.FLAG_BLD and debug.logger(
                        "bad magic in %s" % pycPath
                    )

        for pySfx in SOURCE_SUFFIXES:
            try:
                pyTime = self._get_timestamp(f + pySfx)

            except OSError:
                why = sys.exc_info()[1]
                if ENOENT == -1 or why.errno == ENOENT:
                    debug.logger & debug.FLAG_BLD and debug.logger(
                        f"file {f + pySfx} access error: {why}"
                    )

                else:
                    raise error.MibLoadError(
                        f"MIB file {f + pySfx} access error: {why}"
                    )

            else:
                debug.logger & debug.FLAG_BLD and debug.logger(
                    "file %s mtime %d" % (f + pySfx, pyTime)
                )
                break

        if pycTime != -1 and pycTime >= pyTime:
            return marshal.loads(pycData), pycSfx

        if pyTime != -1:
            modData, pyPath = self._get_data(f + pySfx, "r")
            return compile(modData, pyPath, "exec"), pyPath

        raise OSError(ENOENT, "No suitable module found", f)

    # Interfaces for subclasses
    def _init(self):
        raise NotImplementedError()

    def _listdir(self):
        raise NotImplementedError()

    def _get_timestamp(self, f):
        raise NotImplementedError()

    def _get_data(self, f, mode):
        NotImplementedError()


class ZipMibSource(__AbstractMibSource):
    """Zip MIB source."""

    def _init(self):
        try:
            p = __import__(self._srcName, globals(), locals(), ["__init__"])
            if hasattr(p, "__loader__") and hasattr(p.__loader__, "_files"):
                self.__loader = p.__loader__
                self._srcName = self._srcName.replace(".", os.sep)
                return self
            elif hasattr(p, "__file__"):
                # Dir relative to PYTHONPATH
                return DirMibSource(os.path.split(p.__file__)[0]).init()
            else:
                raise error.MibLoadError(f"{p} access error")

        except ImportError:
            # Dir relative to CWD
            return DirMibSource(self._srcName).init()

    @staticmethod
    def _parse_dos_ime(dos_date, dos_time):
        t = (
            ((dos_date >> 9) & 0x7F) + 1980,  # year
            ((dos_date >> 5) & 0x0F),  # month
            dos_date & 0x1F,  # mday
            (dos_time >> 11) & 0x1F,  # hour
            (dos_time >> 5) & 0x3F,  # min
            (dos_time & 0x1F) * 2,  # sec
            -1,  # wday
            -1,  # yday
            -1,
        )  # dst
        return time.mktime(t)

    def _listdir(self):
        value = []
        # noinspection PyProtectedMember
        for f in self.__loader._files.keys():
            d, f = os.path.split(f)
            if d == self._srcName:
                value.append(f)
        return tuple(self._unique_names(value))

    def _get_timestamp(self, f):
        p = os.path.join(self._srcName, f)
        # noinspection PyProtectedMember
        if p in self.__loader._files:
            # noinspection PyProtectedMember
            return self._parse_dos_ime(
                self.__loader._files[p][6], self.__loader._files[p][5]
            )
        else:
            raise OSError(ENOENT, "No such file in ZIP archive", p)

    def _get_data(self, f, mode=None):
        p = os.path.join(self._srcName, f)
        try:
            return self.__loader.get_data(p), p

        except Exception:  # ZIP code seems to return all kinds of errors
            why = sys.exc_info()
            raise OSError(ENOENT, f"File or ZIP archive {p} access error: {why[1]}")


class DirMibSource(__AbstractMibSource):
    """Directory MIB source."""

    def _init(self):
        self._srcName = os.path.normpath(self._srcName)
        return self

    def _listdir(self):
        try:
            return self._unique_names(os.listdir(self._srcName))
        except OSError:
            why = sys.exc_info()
            debug.logger & debug.FLAG_BLD and debug.logger(
                f"listdir() failed for {self._srcName}: {why[1]}"
            )
            return ()

    def _get_timestamp(self, f):
        p = os.path.join(self._srcName, f)
        try:
            return os.stat(p)[8]
        except OSError:
            raise OSError(ENOENT, "No such file: %s" % sys.exc_info()[1], p)

    def _get_data(self, f, mode):
        p = os.path.join(self._srcName, "*")
        try:
            if f in os.listdir(self._srcName):  # make FS case-sensitive
                p = os.path.join(self._srcName, f)
                fp = open(p, mode)
                data = fp.read()
                fp.close()
                root, extension = os.path.splitext(p)
                return data, extension

        except OSError:
            why = sys.exc_info()
            msg = f"File or directory {p} access error: {why[1]}"

        else:
            msg = "No such file or directory: %s" % p

        raise OSError(ENOENT, msg)


class MibBuilder:
    """MIB builder."""

    DEFAULT_CORE_MIBS = os.pathsep.join(
        ("pysnmp.smi.mibs.instances", "pysnmp.smi.mibs")
    )
    DEFAULT_MISC_MIBS = "pysnmp_mibs"

    module_id = "PYSNMP_MODULE_ID"

    loadTexts = False  # noqa: N815

    # MIB modules can use this to select the features they can use
    version = pysnmp_version

    __mib_sources: list["ZipMibSource | DirMibSource"]

    def __init__(self):
        """Create a MIB builder instance."""
        self.lastBuildId = self._autoName = 0
        sources = []
        for ev in "PYSNMP_MIB_PKGS", "PYSNMP_MIB_DIRS", "PYSNMP_MIB_DIR":
            if ev in os.environ:
                for m in os.environ[ev].split(os.pathsep):
                    sources.append(ZipMibSource(m))
        if not sources and self.DEFAULT_MISC_MIBS:
            for m in self.DEFAULT_MISC_MIBS.split(os.pathsep):
                sources.append(ZipMibSource(m))
        for m in self.DEFAULT_CORE_MIBS.split(os.pathsep):
            sources.insert(0, ZipMibSource(m))
        self.mibSymbols = {}
        self.__mib_sources = []
        self.__modSeen = {}
        self.__modPathsSeen = set()
        self.__mibCompiler = None
        self.set_mib_sources(*sources)

    # MIB compiler management

    def get_mib_compiler(self):
        """Return MIB compiler."""
        return self.__mibCompiler

    def set_mib_compiler(self, mibCompiler, destDir):
        """Set MIB compiler."""
        self.add_mib_sources(DirMibSource(destDir))
        self.__mibCompiler = mibCompiler
        return self

    # MIB modules management

    def add_mib_sources(self, *mibSources: "ZipMibSource | DirMibSource"):
        """Add MIB sources to the search path."""
        self.__mib_sources.extend([s.init() for s in mibSources])
        debug.logger & debug.FLAG_BLD and debug.logger(
            f"addMibSources: new MIB sources {self.__mib_sources}"
        )

    def set_mib_sources(self, *mibSources: "ZipMibSource | DirMibSource"):
        """Set MIB sources to the search path."""
        self.__mib_sources = [s.init() for s in mibSources]
        debug.logger & debug.FLAG_BLD and debug.logger(
            f"setMibSources: new MIB sources {self.__mib_sources}"
        )

    def get_mib_sources(self) -> tuple["ZipMibSource | DirMibSource", ...]:
        """Get MIB sources from the search path."""
        return tuple(self.__mib_sources)

    def load_module(self, modName, **userCtx):
        """Load and execute MIB modules as Python code."""
        for mibSource in self.__mib_sources:
            debug.logger & debug.FLAG_BLD and debug.logger(
                f"loadModule: trying {modName} at {mibSource}"
            )
            try:
                codeObj, sfx = mibSource.read(modName)

            except OSError:
                debug.logger & debug.FLAG_BLD and debug.logger(
                    f"loadModule: read {modName} from {mibSource} failed: {sys.exc_info()[1]}"
                )
                continue

            modPath = mibSource.full_path(modName, sfx)

            if modPath in self.__modPathsSeen:
                debug.logger & debug.FLAG_BLD and debug.logger(
                    "loadModule: seen %s" % modPath
                )
                break

            else:
                self.__modPathsSeen.add(modPath)

            debug.logger & debug.FLAG_BLD and debug.logger(
                "loadModule: evaluating %s" % modPath
            )

            g = {"mibBuilder": self, "userCtx": userCtx}

            try:
                if __debug__:
                    runpy.run_path(
                        modPath, g
                    )  # IMPORTANT: enable break points in loaded MIBs
                else:
                    exec(codeObj, g)

            except Exception:
                self.__modPathsSeen.remove(modPath)
                raise error.MibLoadError(
                    f"MIB module '{modPath}' load error: {traceback.format_exception(*sys.exc_info())}"
                )

            self.__modSeen[modName] = modPath

            debug.logger & debug.FLAG_BLD and debug.logger(
                "loadModule: loaded %s" % modPath
            )

            break

        if modName not in self.__modSeen:
            raise error.MibNotFoundError(
                'MIB file "{}" not found in search path ({})'.format(
                    modName and modName + ".py[co]",
                    ", ".join([str(x) for x in self.__mib_sources]),
                )
            )

        return self

    def load_modules(self, *modNames, **userCtx):
        """Load (optionally, compiling) pysnmp MIB modules."""
        # Build a list of available modules
        if not modNames:
            modNames = {}
            for mibSource in self.__mib_sources:
                for modName in mibSource.listdir():
                    modNames[modName] = None
            modNames = list(modNames)

        if not modNames:
            raise error.MibNotFoundError(f"No MIB module to load at {self}")

        for modName in modNames:
            try:
                self.load_module(modName, **userCtx)

            except error.MibNotFoundError:
                if not self.__mibCompiler:
                    raise

                debug.logger & debug.FLAG_BLD and debug.logger(
                    "loadModules: calling MIB compiler for %s" % modName
                )
                status = self.__mibCompiler.compile(modName, genTexts=self.loadTexts)
                errs = "; ".join(
                    [
                        hasattr(x, "error") and str(x.error) or x
                        for x in status.values()
                        if x in ("failed", "missing")
                    ]
                )
                if errs:
                    raise error.MibNotFoundError(
                        f"{modName} compilation error(s): {errs}"
                    )

                # compilation succeeded, MIB might load now
                self.load_module(modName, **userCtx)

        return self

    def unload_modules(self, *modNames):
        """Unload MIB modules."""
        if not modNames:
            modNames = list(self.mibSymbols.keys())
        for modName in modNames:
            if modName not in self.mibSymbols:
                raise error.MibNotFoundError(f"No module {modName} at {self}")
            self.unexport_symbols(modName)
            self.__modPathsSeen.remove(self.__modSeen[modName])
            del self.__modSeen[modName]

            debug.logger & debug.FLAG_BLD and debug.logger(
                "unloadModules: %s" % modName
            )

        return self

    def import_symbols(self, modName, *symNames, **userCtx) -> "tuple[Any, ...]":
        """Import MIB symbols."""
        if not modName:
            raise error.SmiError("importSymbols: empty MIB module name")
        r = ()
        for symName in symNames:
            if modName not in self.mibSymbols:
                self.load_modules(modName, **userCtx)
            if modName not in self.mibSymbols:
                raise error.MibNotFoundError(f"No module {modName} loaded at {self}")
            if symName not in self.mibSymbols[modName]:
                raise error.SmiError(f"No symbol {modName}::{symName} at {self}")
            r = r + (self.mibSymbols[modName][symName],)
        return r

    def export_symbols(self, modName, *anonymousSyms, **namedSyms):
        """Export MIB symbols."""
        if modName not in self.mibSymbols:
            self.mibSymbols[modName] = {}
        mibSymbols = self.mibSymbols[modName]

        for symObj in anonymousSyms:
            debug.logger & debug.FLAG_BLD and debug.logger(
                "export_symbols: anonymous symbol %s::__pysnmp_%ld"
                % (modName, self._autoName)
            )
            mibSymbols["__pysnmp_%ld" % self._autoName] = symObj
            self._autoName += 1
        for symName, symObj in namedSyms.items():
            if symName in mibSymbols:
                raise error.SmiError(f"Symbol {symName} already exported at {modName}")

            if symName != self.module_id and not isinstance(symObj, classTypes):
                label = symObj.getLabel()
                if label:
                    symName = label
                else:
                    symObj.setLabel(symName)

            mibSymbols[symName] = symObj

            debug.logger & debug.FLAG_BLD and debug.logger(
                f"export_symbols: symbol {modName}::{symName}"
            )

        self.lastBuildId += 1

    def unexport_symbols(self, modName, *symNames):
        """Unexport MIB symbols."""
        if modName not in self.mibSymbols:
            raise error.SmiError(f"No module {modName} at {self}")
        mibSymbols = self.mibSymbols[modName]
        if not symNames:
            symNames = list(mibSymbols.keys())
        for symName in symNames:
            if symName not in mibSymbols:
                raise error.SmiError(f"No symbol {modName}::{symName} at {self}")
            del mibSymbols[symName]

            debug.logger & debug.FLAG_BLD and debug.logger(
                f"unexport_symbols: symbol {modName}::{symName}"
            )

        if not self.mibSymbols[modName]:
            del self.mibSymbols[modName]

        self.lastBuildId += 1

    # Compatibility API
    deprecated_attributes = {
        "importSymbols": "import_symbols",
        "exportSymbols": "export_symbols",
        "unexportSymbols": "unexport_symbols",
        "loadModules": "load_modules",
        "addMibSources": "add_mib_sources",
    }

    def __getattr__(self, attr):
        """Redirect some attrs access to the OID object to behave alike."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from sys import version
from time import time

from pysnmp import __version__

(MibScalarInstance, TimeTicks) = mibBuilder.import_symbols(
    "SNMPv2-SMI", "MibScalarInstance", "TimeTicks"
)

(
    sysDescr,
    sysObjectID,
    sysUpTime,
    sysContact,
    sysName,
    sysLocation,
    sysServices,
    sysORLastChange,
    snmpInPkts,
    snmpOutPkts,
    snmpInBadVersions,
    snmpInBadCommunityNames,
    snmpInBadCommunityUses,
    snmpInASNParseErrs,
    snmpInTooBigs,
    snmpInNoSuchNames,
    snmpInBadValues,
    snmpInReadOnlys,
    snmpInGenErrs,
    snmpInTotalReqVars,
    snmpInTotalSetVars,
    snmpInGetRequests,
    snmpInGetNexts,
    snmpInSetRequests,
    snmpInGetResponses,
    snmpInTraps,
    snmpOutTooBigs,
    snmpOutNoSuchNames,
    snmpOutBadValues,
    snmpOutGenErrs,
    snmpOutSetRequests,
    snmpOutGetResponses,
    snmpOutTraps,
    snmpEnableAuthenTraps,
    snmpSilentDrops,
    snmpProxyDrops,
    snmpTrapOID,
    coldStart,
    snmpSetSerialNo,
) = mibBuilder.import_symbols(
    "SNMPv2-MIB",
    "sysDescr",
    "sysObjectID",
    "sysUpTime",
    "sysContact",
    "sysName",
    "sysLocation",
    "sysServices",
    "sysORLastChange",
    "snmpInPkts",
    "snmpOutPkts",
    "snmpInBadVersions",
    "snmpInBadCommunityNames",
    "snmpInBadCommunityUses",
    "snmpInASNParseErrs",
    "snmpInTooBigs",
    "snmpInNoSuchNames",
    "snmpInBadValues",
    "snmpInReadOnlys",
    "snmpInGenErrs",
    "snmpInTotalReqVars",
    "snmpInTotalSetVars",
    "snmpInGetRequests",
    "snmpInGetNexts",
    "snmpInSetRequests",
    "snmpInGetResponses",
    "snmpInTraps",
    "snmpOutTooBigs",
    "snmpOutNoSuchNames",
    "snmpOutBadValues",
    "snmpOutGenErrs",
    "snmpOutSetRequests",
    "snmpOutGetResponses",
    "snmpOutTraps",
    "snmpEnableAuthenTraps",
    "snmpSilentDrops",
    "snmpProxyDrops",
    "snmpTrapOID",
    "coldStart",
    "snmpSetSerialNo",
)

__sysDescr = MibScalarInstance(
    sysDescr.name,
    (0,),
    sysDescr.syntax.clone(
        "PySNMP engine version {}, Python {}".format(
            __version__, version.replace("\n", " ").replace("\r", " ")
        )
    ),
)
__sysObjectID = MibScalarInstance(
    sysObjectID.name, (0,), sysObjectID.syntax.clone((1, 3, 6, 1, 4, 1, 20408))
)


class SysUpTime(TimeTicks):
    createdAt = time()

    def clone(self, **kwargs):
        if "value" not in kwargs:
            kwargs["value"] = int((time() - self.createdAt) * 100)
        return TimeTicks.clone(self, **kwargs)


__sysUpTime = MibScalarInstance(sysUpTime.name, (0,), SysUpTime(0))
__sysContact = MibScalarInstance(sysContact.name, (0,), sysContact.syntax.clone(""))
__sysName = MibScalarInstance(sysName.name, (0,), sysName.syntax.clone(""))
__sysLocation = MibScalarInstance(sysLocation.name, (0,), sysLocation.syntax.clone(""))
__sysServices = MibScalarInstance(sysServices.name, (0,), sysServices.syntax.clone(0))
__sysORLastChange = MibScalarInstance(
    sysORLastChange.name, (0,), sysORLastChange.syntax.clone(0)
)
__snmpInPkts = MibScalarInstance(snmpInPkts.name, (0,), snmpInPkts.syntax.clone(0))
__snmpOutPkts = MibScalarInstance(snmpOutPkts.name, (0,), snmpOutPkts.syntax.clone(0))
__snmpInBadVersions = MibScalarInstance(
    snmpInBadVersions.name, (0,), snmpInBadVersions.syntax.clone(0)
)
__snmpInBadCommunityNames = MibScalarInstance(
    snmpInBadCommunityNames.name, (0,), snmpInBadCommunityNames.syntax.clone(0)
)
__snmpInBadCommunityUses = MibScalarInstance(
    snmpInBadCommunityUses.name, (0,), snmpInBadCommunityUses.syntax.clone(0)
)
__snmpInASNParseErrs = MibScalarInstance(
    snmpInASNParseErrs.name, (0,), snmpInASNParseErrs.syntax.clone(0)
)
__snmpInTooBigs = MibScalarInstance(
    snmpInTooBigs.name, (0,), snmpInTooBigs.syntax.clone(0)
)
__snmpInNoSuchNames = MibScalarInstance(
    snmpInNoSuchNames.name, (0,), snmpInNoSuchNames.syntax.clone(0)
)
__snmpInBadValues = MibScalarInstance(
    snmpInBadValues.name, (0,), snmpInBadValues.syntax.clone(0)
)
__snmpInReadOnlys = MibScalarInstance(
    snmpInReadOnlys.name, (0,), snmpInReadOnlys.syntax.clone(0)
)
__snmpInGenErrs = MibScalarInstance(
    snmpInGenErrs.name, (0,), snmpInGenErrs.syntax.clone(0)
)
__snmpInTotalReqVars = MibScalarInstance(
    snmpInTotalReqVars.name, (0,), snmpInTotalReqVars.syntax.clone(0)
)
__snmpInTotalSetVars = MibScalarInstance(
    snmpInTotalSetVars.name, (0,), snmpInTotalSetVars.syntax.clone(0)
)
__snmpInGetRequests = MibScalarInstance(
    snmpInGetRequests.name, (0,), snmpInGetRequests.syntax.clone(0)
)
__snmpInGetNexts = MibScalarInstance(
    snmpInGetNexts.name, (0,), snmpInGetNexts.syntax.clone(0)
)
__snmpInSetRequests = MibScalarInstance(
    snmpInSetRequests.name, (0,), snmpInSetRequests.syntax.clone(0)
)
__snmpInGetResponses = MibScalarInstance(
    snmpInGetResponses.name, (0,), snmpInGetResponses.syntax.clone(0)
)
__snmpInTraps = MibScalarInstance(snmpInTraps.name, (0,), snmpInTraps.syntax.clone(0))
__snmpOutTooBigs = MibScalarInstance(
    snmpOutTooBigs.name, (0,), snmpOutTooBigs.syntax.clone(0)
)
__snmpOutNoSuchNames = MibScalarInstance(
    snmpOutNoSuchNames.name, (0,), snmpOutNoSuchNames.syntax.clone(0)
)
__snmpOutBadValues = MibScalarInstance(
    snmpOutBadValues.name, (0,), snmpOutBadValues.syntax.clone(0)
)
__snmpOutGenErrs = MibScalarInstance(
    snmpOutGenErrs.name, (0,), snmpOutGenErrs.syntax.clone(0)
)
__snmpOutSetRequests = MibScalarInstance(
    snmpOutSetRequests.name, (0,), snmpOutSetRequests.syntax.clone(0)
)
__snmpOutGetResponses = MibScalarInstance(
    snmpOutGetResponses.name, (0,), snmpOutGetResponses.syntax.clone(0)
)
__snmpOutTraps = MibScalarInstance(
    snmpOutTraps.name, (0,), snmpOutTraps.syntax.clone(0)
)
__snmpEnableAuthenTraps = MibScalarInstance(
    snmpEnableAuthenTraps.name, (0,), snmpEnableAuthenTraps.syntax.clone(1)
)
__snmpSilentDrops = MibScalarInstance(
    snmpSilentDrops.name, (0,), snmpSilentDrops.syntax.clone(0)
)
__snmpProxyDrops = MibScalarInstance(
    snmpProxyDrops.name, (0,), snmpProxyDrops.syntax.clone(0)
)
__snmpTrapOID = MibScalarInstance(
    snmpTrapOID.name, (0,), snmpTrapOID.syntax.clone(coldStart.name)
)
__snmpSetSerialNo = MibScalarInstance(
    snmpSetSerialNo.name, (0,), snmpSetSerialNo.syntax.clone(0)
)

mibBuilder.export_symbols(
    "__SNMPv2-MIB",
    sysDescr=__sysDescr,
    sysObjectID=__sysObjectID,
    sysUpTime=__sysUpTime,
    sysContact=__sysContact,
    sysName=__sysName,
    sysLocation=__sysLocation,
    sysServices=__sysServices,
    sysORLastChange=__sysORLastChange,
    snmpInPkts=__snmpInPkts,
    snmpOutPkts=__snmpOutPkts,
    snmpInBadVersions=__snmpInBadVersions,
    snmpInBadCommunityNames=__snmpInBadCommunityNames,
    snmpInBadCommunityUses=__snmpInBadCommunityUses,
    snmpInASNParseErrs=__snmpInASNParseErrs,
    snmpInTooBigs=__snmpInTooBigs,
    snmpInNoSuchNames=__snmpInNoSuchNames,
    snmpInBadValues=__snmpInBadValues,
    snmpInReadOnlys=__snmpInReadOnlys,
    snmpInGenErrs=__snmpInGenErrs,
    snmpInTotalReqVars=__snmpInTotalReqVars,
    snmpInTotalSetVars=__snmpInTotalSetVars,
    snmpInGetRequests=__snmpInGetRequests,
    snmpInGetNexts=__snmpInGetNexts,
    snmpInSetRequests=__snmpInSetRequests,
    snmpInGetResponses=__snmpInGetResponses,
    snmpInTraps=__snmpInTraps,
    snmpOutTooBigs=__snmpOutTooBigs,
    snmpOutNoSuchNames=__snmpOutNoSuchNames,
    snmpOutBadValues=__snmpOutBadValues,
    snmpOutGenErrs=__snmpOutGenErrs,
    snmpOutSetRequests=__snmpOutSetRequests,
    snmpOutGetResponses=__snmpOutGetResponses,
    snmpOutTraps=__snmpOutTraps,
    snmpEnableAuthenTraps=__snmpEnableAuthenTraps,
    snmpSilentDrops=__snmpSilentDrops,
    snmpProxyDrops=__snmpProxyDrops,
    snmpTrapOID=__snmpTrapOID,
    snmpSetSerialNo=__snmpSetSerialNo,
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp.proto import errind, error


class AbstractAuthenticationService:
    """Abstract base class for SNMPv3 authentication services."""

    SERVICE_ID = None

    def hash_passphrase(self, authKey):
        """Hash authentication key."""
        raise error.ProtocolError(errind.noAuthentication)

    def localize_key(self, authKey, snmpEngineID):
        """Localize authentication key."""
        raise error.ProtocolError(errind.noAuthentication)

    @property
    def digest_length(self):
        """Return length of the digest produced by this service."""
        raise error.ProtocolError(errind.noAuthentication)

    # 7.2.4.1
    def authenticate_outgoing_message(self, authKey, wholeMsg):
        """Authenticate outgoing message."""
        raise error.ProtocolError(errind.noAuthentication)

    # 7.2.4.2
    def authenticate_incoming_message(self, authKey, authParameters, wholeMsg):
        """Authenticate incoming message."""
        raise error.ProtocolError(errind.noAuthentication)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMP-USM-AES-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source file://asn1/SNMP-USM-AES-MIB
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:39 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(snmpPrivProtocols,) = mibBuilder.import_symbols(
    "SNMP-FRAMEWORK-MIB",
    "snmpPrivProtocols")

(ModuleCompliance,
 NotificationGroup) = mibBuilder.import_symbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 iso,
 snmpModules) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "iso",
    "snmpModules")

(DisplayString,
 TextualConvention) = mibBuilder.import_symbols(
    "SNMPv2-TC",
    "DisplayString",
    "TextualConvention")


# MODULE-IDENTITY

snmpUsmAesMIB = ModuleIdentity(
    (1, 3, 6, 1, 6, 3, 20)
)
if mibBuilder.loadTexts:
    snmpUsmAesMIB.setRevisions(
        ("2004-06-14 00:00",)
    )
if mibBuilder.loadTexts:
    snmpUsmAesMIB.setLastUpdated("200406140000Z")
if mibBuilder.loadTexts:
    snmpUsmAesMIB.setOrganization("IETF")
if mibBuilder.loadTexts:
    snmpUsmAesMIB.setContactInfo("Uri Blumenthal Lucent Technologies / Bell Labs 67 Whippany Rd. 14D-318 Whippany, NJ 07981, USA 973-386-2163 uri@bell-labs.com Fabio Maino Andiamo Systems, Inc. 375 East Tasman Drive San Jose, CA 95134, USA 408-853-7530 fmaino@andiamo.com Keith McCloghrie Cisco Systems, Inc. 170 West Tasman Drive San Jose, CA 95134-1706, USA 408-526-5260 kzm@cisco.com")
if mibBuilder.loadTexts:
    snmpUsmAesMIB.setDescription("Definitions of Object Identities needed for the use of AES by SNMP's User-based Security Model. Copyright (C) The Internet Society (2004). This version of this MIB module is part of RFC 3826; see the RFC itself for full legal notices. Supplementary information may be available on http://www.ietf.org/copyrights/ianamib.html.")


# Types definitions


# TEXTUAL-CONVENTIONS



# MIB Managed Objects in the order of their OIDs

_UsmAesCfb128Protocol_ObjectIdentity = ObjectIdentity
usmAesCfb128Protocol = _UsmAesCfb128Protocol_ObjectIdentity(
    (1, 3, 6, 1, 6, 3, 10, 1, 2, 4)
)
if mibBuilder.loadTexts:
    usmAesCfb128Protocol.setStatus("current")
if mibBuilder.loadTexts:
    usmAesCfb128Protocol.setReference("- Specification for the ADVANCED ENCRYPTION STANDARD. Federal Information Processing Standard (FIPS) Publication 197. (November 2001). - Dworkin, M., NIST Recommendation for Block Cipher Modes of Operation, Methods and Techniques. NIST Special Publication 800-38A (December 2001). ")
if mibBuilder.loadTexts:
    usmAesCfb128Protocol.setDescription("The CFB128-AES-128 Privacy Protocol.")

# Managed Objects groups


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "SNMP-USM-AES-MIB",
    **{"usmAesCfb128Protocol": usmAesCfb128Protocol,
       "snmpUsmAesMIB": snmpUsmAesMIB,
       "PYSNMP_MODULE_ID": snmpUsmAesMIB}
)
